from mpf.platforms.interfaces.segment_display_platform_interface import FlashingType
from mpf.tests.MpfTestCase import MpfTestCase, test_config
from unittest.mock import MagicMock, call
from mpf.platforms import p_roc_common, p_roc


class MockDMD():
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.data = None

    def set_data(self, data):
        self.data = data


class MockPinProcModule(MagicMock):
    DriverCount = 256

    EventTypeAccelerometerIRQ = 11
    EventTypeAccelerometerX = 8
    EventTypeAccelerometerY = 9
    EventTypeAccelerometerZ = 10
    EventTypeBurstSwitchClosed = 7
    EventTypeBurstSwitchOpen = 6
    EventTypeDMDFrameDisplayed = 5
    EventTypeSwitchClosedDebounced = 1
    EventTypeSwitchClosedNondebounced = 3
    EventTypeSwitchOpenDebounced = 2
    EventTypeSwitchOpenNondebounced = 4

    MachineTypeCustom = 1
    MachineTypeInvalid = 0
    MachineTypePDB = 7
    MachineTypeSternSAM = 6
    MachineTypeSternWhitestar = 5
    MachineTypeWPC = 3
    MachineTypeWPC95 = 4
    MachineTypeWPCAlphanumeric = 2

    SwitchCount = 255
    SwitchNeverDebounceFirst = 192
    SwitchNeverDebounceLast = 255

    def aux_command_jump(self, number):
        return "jump" + str(number)

    def aux_command_disable(self):
        return "disable"

    def aux_command_output_custom(self, arg1, arg2, arg3, arg4, arg5):
        return "output_custom_" + str(arg1) + "_" + str(arg2) + "_" + str(arg3) + "_" + str(arg4) + "_" + str(arg5)

    def aux_command_delay(self, delay):
        return "delay_" + str(delay)

    def decode(self, machine_type, device_str):
        if machine_type == MockPinProcModule.MachineTypePDB:
            # for PDB it will just return str as int
            return int(device_str)
        elif machine_type == MockPinProcModule.MachineTypeWPC:
            if device_str == "fllm":
                return 38
            elif device_str == "c01":
                return 40
            elif device_str == "sd1":
                return 8
            elif device_str == "sf1":
                return 0
            elif device_str == "s26":
                return 53
            elif device_str == "s20":
                return 47
            elif device_str == "l11":
                return 80
            elif device_str == "g01":
                return 72
            elif device_str == "c02":
                return 9902
            elif device_str == "c23":
                return 9923
            elif device_str == "c24":
                return 9924
            elif device_str == "c25":
                return 9925
            else:
                raise AssertionError("cannot decode {}".format(device_str))

        else:
            raise AssertionError("Unknown Machine Type {}".format(machine_type))


class TestPRoc(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/p_roc/'

    def get_platform(self):
        # no force platform. we are testing p_roc and p_roc + snux
        return False

    def _normalize(self, type_name):
        self._machine_type = type_name
        if type_name == "pdb":
            return MockPinProcModule.MachineTypePDB
        elif type_name == "wpc":
            return MockPinProcModule.MachineTypeWPC
        else:
            raise AssertionError("Invalid type")

    def read_data(self, module, address):
        return self._memory[module][address]

    def wait_for_platform(self):
        self._sync_count += 1
        num = self._sync_count
        result = self.machine.default_platform.run_proc_cmd_sync("_sync", num)
        assert result[0] == "sync"
        assert result[1] == num

    def _mock_loop(self):
        super()._mock_loop()
        self.loop._wait_for_external_executor = True

    def _driver_get_state(self, driver_num):
        polarity = self._machine_type != "wpc"

        return {
            'driverNum': driver_num,
            'outputDriveTime': 0,
            'polarity': polarity,
            'state': False,
            'waitForFirstTimeSlot': False,
            'timeslots': 0,
            'patterOnTime': 0,
            'patterOffTime': 0,
            'patterEnable': False,
            'futureEnable': False
        }

    def _driver_state_pulse(self, driver, milliseconds):
        driver["state"] = 1
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = False
        driver["outputDriveTime"] = milliseconds
        driver["patterOnTime"] = 0
        driver["patterOffTime"] = 0
        driver["patterEnable"] = False
        driver["futureEnable"] = False
        return driver

    def _driver_state_disable(self, driver):
        driver["state"] = 0
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = False
        driver["outputDriveTime"] = 0
        driver["patterOnTime"] = 0
        driver["patterOffTime"] = 0
        driver["patterEnable"] = False
        driver["futureEnable"] = False
        return driver

    def _driver_state_patter(self, driver, millisecondsOn, millisecondsOff, originalOnTime, now):
        driver["state"] = True
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = not now
        driver["outputDriveTime"] = originalOnTime
        driver["patterOnTime"] = millisecondsOn
        driver["patterOffTime"] = millisecondsOff
        driver["patterEnable"] = True
        driver["futureEnable"] = False
        return driver

    def setUp(self):
        self._sync_count = 0
        self._machine_type = None
        self.expected_duration = 2
        p_roc_common.PINPROC_IMPORTED = True
        p_roc_common.pinproc = MockPinProcModule()
        p_roc.pinproc = p_roc_common.pinproc
        self.pinproc = MagicMock(return_value=True)
        self.pinproc.aux_command_disable = MagicMock(return_value="disable")
        self.pinproc.aux_command_delay = MagicMock(return_value="delay")
        self.pinproc.aux_command_output_custom = MagicMock(return_value="output_custom")
        self.pinproc.aux_command_jump = MagicMock(return_value="jump")
        self.pinproc.read_data = self.read_data
        self.pinproc.aux_send_commands = MagicMock(return_value=True)
        self.pinproc.write_data = MagicMock(return_value=True)
        self.pinproc.flush = MagicMock(return_value=True)
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.pinproc.driver_update_group_config = MagicMock(return_value=True)
        self.pinproc.driver_update_global_config = MagicMock(return_value=True)
        self.pinproc.driver_update_state = MagicMock(return_value=True)
        self.pinproc.driver_pulse = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)
        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_disable = MagicMock(return_value=True)
        self.pinproc.reset = MagicMock(return_value=True)
        self.pinproc.get_events = MagicMock(return_value=[])
        self.pinproc.dmd_update_config = MagicMock(return_value=True)
        self.pinproc.driver_get_state = self._driver_get_state

        p_roc_common.pinproc.PinPROC = MagicMock(return_value=self.pinproc)
        p_roc_common.pinproc.normalize_machine_type = self._normalize
        p_roc_common.pinproc.driver_state_pulse = self._driver_state_pulse

        p_roc_common.pinproc.driver_state_pulsed_patter = None

        p_roc_common.pinproc.driver_state_disable = self._driver_state_disable

        p_roc_common.pinproc.driver_state_patter = self._driver_state_patter

        self.pinproc.switch_get_states = MagicMock(return_value=[0, 1] + [0] * 100)
        self.pinproc.read_data = self.read_data

        self._memory = {
            0x00: {         # manager
                0x00: 0,            # chip id
                0x01: 0x00020006,   # version
                0x03: 0x0000,       # dip switches
            }
        }

        self.pinproc.aux_send_commands = MagicMock(return_value=True)
        super().setUp()

        self.pinproc.aux_send_commands.assert_called_with(0, ["disable"] + ["jump0"] * 254)

    def test_platform(self):
        self._test_initial_switches()
        self._test_switches()
        self._test_pulse_and_hold()
        self._test_pdb_matrix_light()
        self._test_alpha_display()
        self._test_allow_enable()
        self._test_hw_rule_pulse()
        self._test_dmd_update()
        self._test_pdb_gi_light()
        self._test_enable_exception()

        # test hardware scan
        info_str = """Firmware Version: 2 Firmware Revision: 6 Hardware Board ID: 0
"""
        self.assertEqual(info_str, self.machine.default_platform.get_info_string())

    def _test_pulse_and_hold(self):
        # make sure polarity is right
        number = self.machine.coils["c_test"].hw_driver.number
        self.pinproc.driver_update_state.assert_has_calls([
            call({'driverNum': number, 'outputDriveTime': 0, 'polarity': True, 'state': False,
                  'waitForFirstTimeSlot': False, 'timeslots': 0, 'patterOnTime': 0, 'patterOffTime': 0,
                  'patterEnable': False, 'futureEnable': False})],
            False
        )

        self.assertEqual("PD-16 Board 1 Bank 1", self.machine.coils["c_test"].hw_driver.get_board_name())
        # pulse coil A1-B1-2
        self.machine.coils["c_test"].pulse()
        self.wait_for_platform()
        # A1-B1-2 -> address 16 + 8 + 2 = 26 in P3-Roc
        # for 23ms (from config)
        number = self.machine.coils["c_test"].hw_driver.number
        self.pinproc.driver_pulse.assert_called_with(
            number, 23)
        assert not self.pinproc.driver_schedule.called

        self.machine.coils["c_direct2_pulse_power"].pulse()
        self.wait_for_platform()
        number = self.machine.coils["c_direct2_pulse_power"].hw_driver.number
        self.pinproc.driver_pulsed_patter.assert_called_with(
            number, 9, 1, 20, True)

    def _test_alpha_display(self):
        self.pinproc.aux_send_commands = MagicMock(return_value=True)
        self.machine.segment_displays["display1"].add_text("1234", key="score")
        self.machine_run()
        self.wait_for_platform()
        self.pinproc.aux_send_commands.assert_has_calls([
            call(0, ['disable', 'output_custom_0_0_8_False_0', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0',
                     'delay_350', 'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_2',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'delay_40',
                     'output_custom_1_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
                     'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350',
                     'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_2',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'delay_40',
                     'output_custom_2_0_8_False_0', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0',
                     'delay_350', 'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_2',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'delay_40',
                     'output_custom_3_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
                     'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350',
                     'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_2',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'delay_40',
                     'output_custom_4_0_8_False_0', 'output_custom_6_0_9_False_0', 'output_custom_0_0_10_False_0',
                     'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350',
                     'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_2',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'delay_40',
                     'output_custom_5_0_8_False_0', 'output_custom_91_0_9_False_0', 'output_custom_8_0_10_False_0',
                     'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350',
                     'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_2',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'delay_40',
                     'output_custom_6_0_8_False_0', 'output_custom_79_0_9_False_0',
                     'output_custom_8_0_10_False_0', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'delay_40', 'output_custom_7_0_8_False_0',
                     'output_custom_102_0_9_False_0', 'output_custom_8_0_10_False_0', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'delay_40', 'output_custom_8_0_8_False_0',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'delay_40', 'output_custom_9_0_8_False_0',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'delay_40', 'output_custom_10_0_8_False_0',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'delay_40', 'output_custom_11_0_8_False_0',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'delay_40', 'output_custom_12_0_8_False_0',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'delay_40', 'output_custom_13_0_8_False_0',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'delay_40', 'output_custom_14_0_8_False_0',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'delay_40', 'output_custom_15_0_8_False_0',
                     'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
                     'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0',
                     'output_custom_0_0_10_False_0', 'delay_40']),
            call(0, ["jump1"])
        ], any_order=False)

        self.pinproc.aux_send_commands = MagicMock(return_value=True)
        self.machine.segment_displays["display1"].set_flashing(FlashingType.FLASH_ALL)
        self.advance_time_and_run(2)
        self.machine_run()
        self.wait_for_platform()
        self.pinproc.aux_send_commands.assert_has_calls([
            call(0, [
            'disable', 'output_custom_0_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_1_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_2_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_3_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_4_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_5_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_6_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_7_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_8_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_9_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_10_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_11_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_12_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_13_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_14_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40', 'output_custom_15_0_8_False_0', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'output_custom_0_0_11_False_0', 'output_custom_0_0_12_False_0', 'delay_350', 'output_custom_0_0_11_False_0',
            'output_custom_0_0_12_False_0', 'delay_2', 'output_custom_0_0_9_False_0', 'output_custom_0_0_10_False_0',
            'delay_40']),
            call(0, ["jump1"])
        ], any_order=True)

    def _test_enable_exception(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.coils["c_test"].enable()

    def _test_allow_enable(self):
        self.machine.coils["c_test_allow_enable"].enable()
        self.wait_for_platform()
        number = self.machine.coils["c_test_allow_enable"].hw_driver.number
        self.pinproc.driver_schedule.assert_called_with(
            number, 0xffffffff, 0, True)

    def _test_hw_rule_pulse(self):
        self.pinproc.switch_update_rule = MagicMock()
        self.machine.autofires["ac_slingshot_test"].enable()
        self.wait_for_platform()
        coil_number = self.machine.coils["c_slingshot_test"].hw_driver.number
        self.pinproc.switch_update_rule.assert_has_calls([
            call(40, 'open_nondebounced', {'notifyHost': False, 'reloadActive': True}, [], False),
            call(40, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': True},
                 [{'futureEnable': False, 'patterOffTime': 0, 'polarity': True, 'waitForFirstTimeSlot': False,
                   'timeslots': 0, 'patterOnTime': 0, 'outputDriveTime': 10, 'patterEnable': False, 'state': 1,
                   'driverNum': coil_number}], False),
            call(40, 'open_debounced', {'notifyHost': True, 'reloadActive': True}, [], False),
            call(40, 'closed_debounced', {'notifyHost': True, 'reloadActive': True}, [], False),
        ], any_order=True)
        self.machine.autofires["ac_slingshot_test"].disable()

    def _test_initial_switches(self):
        self.assertMachineVarEqual(2, "p_roc_version")
        self.assertMachineVarEqual(6, "p_roc_revision")

        self.assertEqual(0x2, self.machine.default_platform.version)
        self.assertEqual(0x6, self.machine.default_platform.revision)

        self.assertSwitchState("s_test", 0)
        self.assertSwitchState("s_test_000", 0)
        self.assertSwitchState("s_direct", 1)

    def _test_switches(self):
        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(23, 'closed_debounced', {'notifyHost': True, 'reloadActive': False}, [], False),
            call(23, 'open_debounced', {'notifyHost': True, 'reloadActive': False}, [], False),
            call(24, 'closed_nondebounced', {'notifyHost': True, 'reloadActive': False}, [], False),
            call(24, 'open_nondebounced', {'notifyHost': True, 'reloadActive': False}, [], False),
        ], any_order=True)

        self.assertSwitchState("s_test", 0)
        # closed debounced -> switch active
        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 1, 'value': 23}])
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.assertSwitchState("s_test", 1)

        # open debounces -> inactive
        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 2, 'value': 23}])
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.assertSwitchState("s_test", 0)

        self.assertSwitchState("s_test_no_debounce", 0)
        # closed non debounced -> should be active
        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 3, 'value': 24}])
        self.wait_for_platform()
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.assertSwitchState("s_test_no_debounce", 1)

        # open non debounced -> should be inactive
        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 4, 'value': 24}])
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.assertSwitchState("s_test_no_debounce", 0)

        # restore empty handler
        self.pinproc.get_events = MagicMock(return_value=[])

    def _test_dmd_update(self):
        self.machine.default_platform.pinproc.DMDBuffer = MockDMD
        self.pinproc.dmd_draw = MagicMock(return_value=True)

        # test configure
        self.machine.default_platform.configure_dmd()
        self.wait_for_platform()

        self.pinproc.dmd_update_config.assert_called_with([1, 2, 3, 4])

        # test set frame to buffer
        frame = bytearray()
        for i in range(4096):
            frame.append(i % 256)

        dmd = self.machine.default_platform.dmd

        dmd.update(frame)
        self.wait_for_platform()

        self.assertEqual(self.pinproc.dmd_draw.call_args[0][0].data, frame)

        # frame displayed
        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 5, 'value': 123}])

        self.advance_time_and_run(0.04)
        self.wait_for_platform()
        self.assertEqual(self.pinproc.dmd_draw.call_args[0][0].data, frame)

        self.pinproc.dmd_draw = MagicMock(return_value=True)

        # test set frame to buffer
        frame = bytearray()
        for i in range(1234):
            frame.append(i % 256)

        dmd.update(frame)
        self.wait_for_platform()

        # should not be rendered
        assert not self.pinproc.dmd_draw.called

        # restore empty handler
        self.pinproc.get_events = MagicMock(return_value=[])

    def _test_pdb_matrix_light(self):
        # very simple check for matrix config
        self.pinproc.driver_update_group_config.assert_has_calls(
            [call(4, 100, 5, 0, 0, True, True, True, True)]
        )

        # test enable of matrix light
        assert not self.pinproc.driver_patter.called
        assert not self.pinproc.driver_schedule.called
        self.machine.lights["test_pdb_light"].on()
        self.advance_time_and_run(.02)
        self.wait_for_platform()
        self.pinproc.driver_schedule.assert_called_with(
            32, 4294967295, 0, True
        )

        self.machine.lights["test_pdb_light"].on(brightness=128)
        self.advance_time_and_run(.02)
        self.wait_for_platform()
        self.pinproc.driver_patter.assert_called_with(
            32, 1, 1, 0, True
        )

        # test disable of matrix light
        assert not self.pinproc.driver_disable.called
        self.machine.lights["test_pdb_light"].off()
        self.machine_run()
        self.wait_for_platform()
        self.pinproc.driver_disable.assert_called_with(32)

        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)
        self.pinproc.driver_disable = MagicMock(return_value=True)

        self.machine.lights["test_pdb_light"].on()
        self.wait_for_platform()
        self.pinproc.driver_schedule.assert_called_with(
            32, 4294967295, 0, True
        )
        self.pinproc.driver_schedule = MagicMock(return_value=True)

    def _test_pdb_gi_light(self):
        # test gi on
        device = self.machine.lights["test_gi"]
        num = self.machine.coils["test_gi"].hw_driver.number
        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)
        device.color("white")
        self.machine_run()
        self.wait_for_platform()
        self.pinproc.driver_schedule.assert_has_calls([
            call(num, 4294967295, 0, True)])
        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)

        device.color([128, 128, 128])
        self.machine_run()
        self.wait_for_platform()
        self.pinproc.driver_patter.assert_has_calls([
            call(num, 1, 1, 0, True)])
        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)

        device.color([245, 245, 245])
        self.machine_run()
        self.wait_for_platform()
        self.pinproc.driver_patter.assert_has_calls([
            call(num, 19, 1, 0, True)])
        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)

        # test gi off
        self.pinproc.driver_disable = MagicMock(return_value=True)
        device.color("off")
        self.machine_run()
        self.wait_for_platform()
        self.pinproc.driver_disable.assert_has_calls([
            call(num)])

    @test_config("wpc.yaml")
    def test_load_wpc(self):
        # make sure p-roc properly initialises with WPC config

        # make sure polarity is right
        number = self.machine.coils["c_test_direct"].hw_driver.number
        self.pinproc.driver_update_state.assert_has_calls([
            call({'driverNum': number, 'outputDriveTime': 0, 'polarity': False, 'state': False,
                  'waitForFirstTimeSlot': False, 'timeslots': 0, 'patterOnTime': 0, 'patterOffTime': 0,
                  'patterEnable': False, 'futureEnable': False})],
            False
        )

        # test polarity in rule
        self.pinproc.switch_update_rule = MagicMock()
        self.machine.autofires["ac_slingshot_test"].enable()
        self.wait_for_platform()
        coil_number = self.machine.coils["c_slingshot_test"].hw_driver.number
        self.pinproc.switch_update_rule.assert_has_calls([
            call(47, 'open_nondebounced', {'notifyHost': False, 'reloadActive': True}, [], False),
            call(47, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': True},
                 [{'futureEnable': False, 'patterOffTime': 0, 'polarity': False, 'waitForFirstTimeSlot': False,
                   'timeslots': 0, 'patterOnTime': 0, 'outputDriveTime': 10, 'patterEnable': False, 'state': 1,
                   'driverNum': coil_number}], False),
            call(47, 'open_debounced', {'notifyHost': True, 'reloadActive': True}, [], False),
            call(47, 'closed_debounced', {'notifyHost': True, 'reloadActive': True}, [], False),
        ], any_order=True)
        self.machine.autofires["ac_slingshot_test"].disable()

    @test_config("snux.yaml")
    def test_load_snux(self):
        """Test snux."""
        # test enable
        self.machine.coils["c_flipper_enable_driver"].enable()
        self.wait_for_platform()
        self.pinproc.driver_schedule.assert_called_with(
            9923, 0xffffffff, 0, True)

        # assert diag flash
        self.advance_time_and_run(1)
        self.wait_for_platform()
        self.pinproc.driver_pulse.assert_called_with(
            9924, 250)

        self.advance_time_and_run()
        self.wait_for_platform()

        # pulse a and c side
        self.pinproc.driver_schedule = MagicMock(return_value=True)
        self.pinproc.driver_pulse = MagicMock(return_value=True)
        self.machine.coils["c_test_a_side"].pulse(100)
        self.advance_time_and_run(.001)
        self.machine.coils["c_test_c_side"].pulse(50)
        self.advance_time_and_run(.040)
        self.wait_for_platform()
        self.advance_time_and_run(.001)
        self.wait_for_platform()
        self.pinproc.driver_pulse.assert_has_calls([
            call(9902, 100)]
        )
        self.assertFalse(self.pinproc.driver_schedule.called)
        self.pinproc.driver_pulse = MagicMock(return_value=True)

        # afterwards service c side
        self.advance_time_and_run(.2)
        self.wait_for_platform()
        self.pinproc.driver_schedule.assert_called_with(
            9925, 0xffffffff, 0, True)
        self.pinproc.driver_pulse.assert_has_calls([
            call(9902, 50)]
        )



"""Contains code for SmartMatrix RGB DMD."""

import logging

import threading
from typing import Dict
import serial

from mpf.platforms.interfaces.dmd_platform import DmdPlatformInterface

from mpf.exceptions.config_file_error import ConfigFileError

from mpf.core.platform import RgbDmdPlatform
from mpf.core.utility_functions import Util


class SmartMatrixHardwarePlatform(RgbDmdPlatform):

    """SmartMatrix RGB DMD."""

    __slots__ = ["devices"]

    def __init__(self, machine):
        """Initialise RGB DMD."""
        super().__init__(machine)
        self.features['tickless'] = True

        self.configure_logging('SmartMatrix')
        self.log.info("Configuring SmartMatrix RGB DMD hardware interface.")

        self.devices = dict()       # type: Dict[str, SmartMatrixDevice]

        if not isinstance(self.machine.config['smartmatrix'], dict):
            raise ConfigFileError("Smartmatrix config needs to be a dict.", 1, self.log.name)

        for name, config in self.machine.config['smartmatrix'].items():
            config = self.machine.config_validator.validate_config(
                config_spec='smartmatrix',
                source=config)
            self.devices[name] = SmartMatrixDevice(config, machine)

    async def initialize(self):
        """Initialise platform."""
        for device in self.devices.values():
            await device.connect()

    def stop(self):
        """Stop platform."""
        for device in self.devices.values():
            device.stop()

        self.devices = {}

    def __repr__(self):
        """Return string representation."""
        return '<Platform.SmartMatrix>'

    def configure_rgb_dmd(self, name: str):
        """Configure rgb dmd."""
        return self.devices[name]


class SmartMatrixDevice(DmdPlatformInterface):

    """A smartmatrix device."""

    __slots__ = ["config", "writer", "port", "control_data_queue", "current_frame", "new_frame_event", "machine", "log"]

    def __init__(self, config, machine):
        """Initialise smart matrix device."""
        self.config = config
        self.writer = None
        self.port = None
        self.control_data_queue = None
        self.current_frame = None
        self.new_frame_event = None
        self.machine = machine
        self.log = logging.getLogger('SmartMatrixDevice')

    def _feed_hardware(self):
        """Feed hardware in separate thread.

        Wait for new_frame_event and send the last frame. If no event happened for 1s refresh the last frame.
        """
        while not self.machine.thread_stopper.is_set():
            # wait for new frame or timeout
            self.new_frame_event.wait(1)

            # clear event
            self.new_frame_event.clear()

            # check if we need to send any control data
            while self.control_data_queue:
                self.port.write(self.control_data_queue.pop())

            # do not crash on missing frame
            if self.current_frame is None:
                continue

            # send frame
            if self.config['old_cookie']:
                self.port.write(bytearray([0x01]) + self.current_frame)
            else:
                self.port.write(bytearray([0xBA, 0x11, 0x00, 0x03, 0x04, 0x00, 0x00, 0x00]) + self.current_frame)

        # close port before exit
        self.port.close()

    async def connect(self):
        """Connect to SmartMatrix device."""
        self.log.info("Connecting to SmartMatrix RGB DMD on %s baud %s", self.config['port'], self.config['baud'])
        self.port = serial.Serial(self.config['port'], self.config['baud'])
        self.new_frame_event = threading.Event()
        self.control_data_queue = []
        self.writer = self.machine.clock.loop.run_in_executor(None, self._feed_hardware)
        self.writer.add_done_callback(Util.raise_exceptions)

    def set_brightness(self, brightness: float):
        """Set brightness."""
        if brightness < 0.0 or brightness > 1.0:
            raise AssertionError("Brightness has to be between 0 and 1.")
        if not self.config['old_cookie']:
            self.control_data_queue.insert(0, bytearray([0xBA, 0x11, 0x00, 0x03, 20, int(brightness * 255), 00, 00]))

    def stop(self):
        """Stop platform."""

    def update(self, data):
        """Update DMD data."""
        self.current_frame = bytearray(data)
        self.new_frame_event.set()



"""Bonus mode for MPF."""
from mpf.core.mode import Mode


class Bonus(Mode):

    """Bonus mode for MPF.

    Give a player bonus for their achievements, but only if the machine is not
    tilted.
    """

    __slots__ = ["bonus_score", "settings", "display_delay", "bonus_entries", "bonus_iterator"]

    def __init__(self, *args, **kwargs):
        """Initialize bonus mode."""
        super().__init__(*args, **kwargs)
        self.bonus_score = None

        self.settings = self.machine.config_validator.validate_config(
            'bonus_mode_settings', self.config.get("mode_settings"))

        self.display_delay = self.settings["display_delay_ms"]
        self.bonus_entries = self.settings["bonus_entries"]
        self.bonus_iterator = None

    def mode_start(self, **kwargs):
        """Start the bonus mode and setup all handlers."""
        if not self.bonus_entries:
            raise ValueError(
                "Bonus mode started, but `bonus_entries` is not configured.")

        if self.machine.game.tilted:
            self.debug_log("Machine is tilted. Skipping bonus.")
            self._reset_all_scores()
            self.stop()
            return

        self.bonus_score = 0
        self.bonus_iterator = iter(self.bonus_entries)
        self.display_delay = self.settings["display_delay_ms"]
        self.machine.events.post('bonus_start')
        '''event: bonus_start

        desc: The end-of-ball bonus is starting. You can use this event in
        your slide player to trigger the bonus intro slide. If the game has
        tilted, this event will not be posted.

        '''
        self.delay.add(name='bonus', ms=self.display_delay,
                       callback=self._bonus_next_item)

        if self.settings["hurry_up_event"]:
            self.add_mode_event_handler(self.settings['hurry_up_event'],
                                        self.hurry_up)

        if self.settings["end_bonus_event"]:
            self.add_mode_event_handler(self.settings['end_bonus_event'],
                                        self._end_bonus)

    def hurry_up(self, **kwargs):
        """Change the slide display delay to the "hurry up" setting.

        This is typically used with a flipper cancel event to hurry up the
        bonus display when the player hits both flippers.
        """
        del kwargs
        self.display_delay = self.settings["hurry_up_delay_ms"]
        self.delay.run_now('bonus')

    def _reset_all_scores(self):
        """Reset score entries without scoring them.

        We keep the permanent entries.
        """
        self.debug_log("Resetting player_score_entries that are set to reset.")
        for entry in self.bonus_entries:
            if entry['reset_player_score_entry']:
                self.player.vars[entry['player_score_entry']] = 0

    def _bonus_next_item(self):

        try:
            entry = next(self.bonus_iterator)
        except StopIteration:
            self._subtotal()
            return

        hits = self.player.vars.get(entry['player_score_entry'], 1)
        score = entry['score'].evaluate([]) * hits

        if not score and entry['skip_if_zero']:
            self.debug_log("Skipping bonus entry '{}' because its value is 0".
                           format(entry['event']))
            self._bonus_next_item()
            return

        self.debug_log("Bonus Entry '{}': score: {} player_score_entry: {}={}".
                       format(entry['event'], score,
                              entry['player_score_entry'], hits))

        self.bonus_score += score
        self.machine.events.post(entry['event'], score=score,
                                 bonus_score=self.bonus_score, hits=hits)
        if entry['reset_player_score_entry']:
            self.player.vars[entry['player_score_entry']] = 0

        self.delay.add(name='bonus', ms=self.display_delay,
                       callback=self._bonus_next_item)

    def _subtotal(self):
        if self.player.vars.get("bonus_multiplier", 1) == 1:
            self.debug_log(
                "Skipping bonus_multiplier event because the multiplier is 1.")
            self._total_bonus()

        else:
            self.debug_log("Bonus subtotal: {}", self.bonus_score)
            self.machine.events.post('bonus_subtotal', score=self.bonus_score)
            '''event: bonus_subtotal

            desc: Posted by the bonus mode after all the individual bonus
            entries have been posted and processed.

            This event is typically posted just before the bonus multiplier
            screen, so if the bonus multiplier is 1, then this event will
            be skipped.

            args:

            score: The score of the bonus (so far)

            '''
            self.delay.add(name='bonus', ms=self.display_delay,
                           callback=self._do_multiplier)

    def _do_multiplier(self):
        multiplier = self.player.vars.get("bonus_multiplier", 1)
        self.debug_log("Bonus multiplier: {}".format(multiplier))
        self.machine.events.post('bonus_multiplier', multiplier=multiplier)
        '''event: bonus_multiplier

        desc: Posted after "bonus_subtotal" and used to trigger the bonus
        multiplier screen. If the bonus multiplier is 1, then this event is
        skipped.

        args:

        multiplier: The numeric value of the bonus multiplier.

        '''
        self.bonus_score *= multiplier
        self.delay.add(name='bonus', ms=self.display_delay,
                       callback=self._total_bonus)

    def _total_bonus(self):
        self.player.score += self.bonus_score
        self.debug_log("Bonus Total: {}", self.bonus_score)
        self.machine.events.post('bonus_total', score=self.bonus_score)

        if not self.settings['end_bonus_event']:
            self.delay.add(name='bonus', ms=self.display_delay,
                           callback=self._end_bonus)

    def _end_bonus(self, **kwargs):
        del kwargs
        self.debug_log("Bonus done")
        keep_multiplier = self.settings['keep_multiplier']

        if not keep_multiplier.evaluate({}):
            self.player.bonus_multiplier = 1
        self.stop()



"""Command dispatcher of the Mission Pinball Framework."""
import sys
import mpf.commands


def main(args=None):
    """Dispatche commands to our handlers."""
    if args is None:
        args = sys.argv[1:]

    mpf.commands.run_from_command_line(args)


if __name__ == "__main__":
    main()



"""Enable/disable ejector."""
from mpf.devices.ball_device.ball_device_ejector import BallDeviceEjector

from mpf.core.delays import DelayManager
from mpf.devices.ball_device.default_ball_search import DefaultBallSearch


class EnableCoilEjector(DefaultBallSearch, BallDeviceEjector):

    """Enable a coil to eject one ball."""

    __slots__ = ["delay"]

    def __init__(self, config, ball_device, machine):
        """Initialise ejector."""
        for option in ["eject_coil", "eject_coil_enable_time"]:
            if option not in config and option in ball_device.config:
                config[option] = ball_device.config[option]

        super().__init__(config, ball_device, machine)
        self.delay = DelayManager(self.ball_device.machine)

        self.config = self.machine.config_validator.validate_config("ball_device_ejector_enable", self.config)

    async def eject_one_ball(self, is_jammed, eject_try, balls_in_device):
        """Enable eject coil."""
        del is_jammed
        del eject_try
        # If multiple eject_coil_enable_time values, they correspond to the # of balls
        if self.ball_device.balls <= len(self.config['eject_coil_enable_time']):
            eject_time = self.config['eject_coil_enable_time'][balls_in_device - 1]
        else:
            eject_time = self.config['eject_coil_enable_time'][-1]

        # default pulse
        self.ball_device.debug_log("Enabling eject coil for %sms, Current balls: %s.",
                                   eject_time,
                                   self.ball_device.balls)

        self.config['eject_coil'].enable(max_wait_ms=self.config['eject_coil_max_wait_ms'])
        self.delay.reset(name="disable", callback=self._disable_coil,
                         ms=eject_time)

    async def reorder_balls(self):
        """Reordering balls is not supported."""
        # TODO: implement
        self.ball_device.log.warning("Reordering balls is not implemented in enable ejector")

    def _disable_coil(self):
        """Disable the coil."""
        self.config['eject_coil'].disable()

    def _fire_coil_for_search(self, full_power):
        if not full_power:
            self.config['eject_coil'].pulse()
        else:
            self.config['eject_coil'].enable()
            self.delay.reset(name="disable", callback=self._disable_coil,
                             ms=self.config['eject_coil_enable_time'][0])
        return True



# pylint: disable-msg=too-many-lines
"""OPP Hardware interface.

Contains the hardware interface and drivers for the Open Pinball Project
platform hardware, including the solenoid, input, incandescent, and neopixel
boards.
"""
import asyncio
from collections import defaultdict
from typing import Dict, List, Set, Union, Tuple, Optional  # pylint: disable-msg=cyclic-import,unused-import

from mpf.core.platform_batch_light_system import PlatformBatchLightSystem
from mpf.core.utility_functions import Util

from mpf.platforms.interfaces.driver_platform_interface import PulseSettings, HoldSettings

from mpf.platforms.opp.opp_coil import OPPSolenoidCard
from mpf.platforms.opp.opp_incand import OPPIncandCard
from mpf.platforms.opp.opp_modern_lights import OPPModernLightChannel, OPPNeopixelCard, OPPModernMatrixLightsCard
from mpf.platforms.opp.opp_serial_communicator import OPPSerialCommunicator, BAD_FW_VERSION
from mpf.platforms.opp.opp_switch import OPPInputCard
from mpf.platforms.opp.opp_switch import OPPMatrixCard
from mpf.platforms.opp.opp_rs232_intf import OppRs232Intf
from mpf.core.platform import SwitchPlatform, DriverPlatform, LightsPlatform, SwitchSettings, DriverSettings, \
    DriverConfig, SwitchConfig, RepulseSettings

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.platforms.opp.opp_coil import OPPSolenoid  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.platforms.opp.opp_incand import OPPIncand  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.platforms.opp.opp_switch import OPPSwitch  # pylint: disable-msg=cyclic-import,unused-import


# pylint: disable-msg=too-many-instance-attributes
class OppHardwarePlatform(LightsPlatform, SwitchPlatform, DriverPlatform):

    """Platform class for the OPP hardware.

    Args:
    ----
        machine: The main ``MachineController`` instance.

    """

    __slots__ = ["opp_connection", "serial_connections", "opp_incands", "opp_solenoid", "sol_dict",
                 "opp_inputs", "inp_dict", "inp_addr_dict", "matrix_inp_addr_dict", "read_input_msg",
                 "neo_card_dict", "num_gen2_brd", "gen2_addr_arr", "bad_crc", "min_version", "_poll_task",
                 "config", "_poll_response_received", "machine_type", "opp_commands", "_incand_task", "_light_system",
                 "matrix_light_cards"]

    def __init__(self, machine) -> None:
        """Initialise OPP platform."""
        super().__init__(machine)
        self.opp_connection = {}            # type: Dict[str, OPPSerialCommunicator]
        self.serial_connections = set()     # type: Set[OPPSerialCommunicator]
        self.opp_incands = dict()           # type: Dict[str, OPPIncandCard]
        self.opp_solenoid = []              # type: List[OPPSolenoidCard]
        self.sol_dict = dict()              # type: Dict[str, OPPSolenoid]
        self.opp_inputs = []                # type: List[Union[OPPInputCard, OPPMatrixCard]]
        self.inp_dict = dict()              # type: Dict[str, OPPSwitch]
        self.inp_addr_dict = dict()         # type: Dict[str, OPPInputCard]
        self.matrix_inp_addr_dict = dict()  # type: Dict[str, OPPMatrixCard]
        self.read_input_msg = {}            # type: Dict[str, bytes]
        self.neo_card_dict = dict()         # type: Dict[str, OPPNeopixelCard]
        self.matrix_light_cards = dict()    # type: Dict[str, OPPModernMatrixLightsCard]
        self.num_gen2_brd = 0
        self.gen2_addr_arr = {}             # type: Dict[str, Dict[int, Optional[int]]]
        self.bad_crc = 0
        self.min_version = defaultdict(lambda: 0xffffffff)      # type: Dict[str, int]
        self._poll_task = {}                # type: Dict[str, asyncio.Task]
        self._incand_task = None            # type: Optional[asyncio.Task]
        self._light_system = None           # type: Optional[PlatformBatchLightSystem]

        self.features['tickless'] = True

        self.config = self.machine.config_validator.validate_config("opp", self.machine.config['opp'])
        self._configure_device_logging_and_debug("OPP", self.config)
        self._poll_response_received = {}   # type: Dict[str, asyncio.Event]
        assert self.log is not None

        if self.config['driverboards']:
            self.machine_type = self.config['driverboards']
        else:
            self.machine_type = self.machine.config['hardware']['driverboards'].lower()

        if self.machine_type == 'gen1':
            raise AssertionError("Original OPP boards not currently supported.")
        if self.machine_type == 'gen2':
            self.debug_log("Configuring the OPP Gen2 boards")
        else:
            self.raise_config_error('Invalid driverboards type: {}'.format(self.machine_type), 15)

        # Only including responses that should be received
        self.opp_commands = {
            ord(OppRs232Intf.INV_CMD): self.inv_resp,
            ord(OppRs232Intf.EOM_CMD): self.eom_resp,
            ord(OppRs232Intf.GET_GEN2_CFG): self.get_gen2_cfg_resp,
            ord(OppRs232Intf.READ_GEN2_INP_CMD): self.read_gen2_inp_resp_initial,
            ord(OppRs232Intf.GET_VERS_CMD): self.vers_resp,
            ord(OppRs232Intf.READ_MATRIX_INP): self.read_matrix_inp_resp_initial,
        }

    async def initialize(self):
        """Initialise connections to OPP hardware."""
        await self._connect_to_hardware()
        self.opp_commands[ord(OppRs232Intf.READ_GEN2_INP_CMD)] = self.read_gen2_inp_resp
        self.opp_commands[ord(OppRs232Intf.READ_MATRIX_INP)] = self.read_matrix_inp_resp

        self._light_system = PlatformBatchLightSystem(self.machine.clock, self._send_multiple_light_update,
                                                      self.machine.config['mpf']['default_light_hw_update_hz'],
                                                      128)

    async def _send_multiple_light_update(self, sequential_brightness_list: List[Tuple[OPPModernLightChannel,
                                                                                       float, int]]):
        first_light, _, common_fade_ms = sequential_brightness_list[0]
        number_leds = len(sequential_brightness_list)

        msg = bytearray()
        msg.append(int(ord(OppRs232Intf.CARD_ID_GEN2_CARD) + first_light.addr))
        msg.append(OppRs232Intf.SERIAL_LED_CMD_FADE)
        msg.append(int(first_light.pixel_num / 256))
        msg.append(int(first_light.pixel_num % 256))
        msg.append(int(number_leds / 256))
        msg.append(int(number_leds % 256))
        msg.append(int(common_fade_ms / 256))
        msg.append(int(common_fade_ms % 256))

        for _, brightness, _ in sequential_brightness_list:
            msg.append(int(brightness * 255))

        msg.extend(OppRs232Intf.calc_crc8_whole_msg(msg))
        cmd = bytes(msg)

        if self.debug:
            self.debug_log("Set color on %s: %s", first_light.chain_serial, "".join(" 0x%02x" % b for b in cmd))
        self.send_to_processor(first_light.chain_serial, cmd)

    async def start(self):
        """Start polling and listening for commands."""
        # start polling
        for chain_serial in self.read_input_msg:
            self._poll_task[chain_serial] = self.machine.clock.loop.create_task(self._poll_sender(chain_serial))
            self._poll_task[chain_serial].add_done_callback(Util.raise_exceptions)

        # start listening for commands
        for connection in self.serial_connections:
            await connection.start_read_loop()

        if [version for version in self.min_version.values() if version < 0x02010000]:
            # if we run any CPUs with firmware prior to 2.1.0 start incands updater
            self._incand_task = self.machine.clock.schedule_interval(self.update_incand,
                                                                     1 / self.config['incand_update_hz'])

        self._light_system.start()

    def stop(self):
        """Stop hardware and close connections."""
        for task in self._poll_task.values():
            task.cancel()

        self._poll_task = {}

        if self._incand_task:
            self._incand_task.cancel()
            self._incand_task = None

        for connections in self.serial_connections:
            connections.stop()

        self.serial_connections = []

    def __repr__(self):
        """Return string representation."""
        return '<Platform.OPP>'

    def process_received_message(self, chain_serial, msg):
        """Send an incoming message from the OPP hardware to the proper method for servicing.

        Args:
        ----
            chain_serial: Serial of the chain which received the message.
            msg: Message to parse.
        """
        if len(msg) >= 1:
            # Verify valid Gen2 address
            if (msg[0] & 0xe0) == 0x20:
                if len(msg) >= 2:
                    cmd = msg[1]
                else:
                    cmd = OppRs232Intf.ILLEGAL_CMD
            # Look for EOM or INV commands
            elif msg[0] == ord(OppRs232Intf.INV_CMD) or msg[0] == ord(OppRs232Intf.EOM_CMD):
                cmd = msg[0]
            else:
                cmd = OppRs232Intf.ILLEGAL_CMD
        else:
            # No messages received, fake an EOM
            cmd = OppRs232Intf.EOM_CMD

        # Can't use try since it swallows too many errors for now
        if cmd in self.opp_commands:
            self.opp_commands[cmd](chain_serial, msg)
        else:
            self.log.warning("Received unknown serial command?%s. (This is "
                             "very worrisome.)", "".join(" 0x%02x" % b for b in msg))

            # TODO: This means synchronization is lost.  Send EOM characters
            #  until they come back
            self.opp_connection[chain_serial].lost_synch()

    @staticmethod
    def _get_numbers(mask):
        number = 0
        ref = 1
        result = []
        while mask > ref:
            if mask & ref:
                result.append(number)
            number += 1
            ref = ref << 1

        return result

    def get_info_string(self):
        """Dump infos about boards."""
        if not self.serial_connections:
            return "No connection to any CPU board."

        infos = "Connected CPUs:\n"
        for connection in sorted(self.serial_connections, key=lambda x: x.chain_serial):
            infos += " - Port: {} at {} baud. Chain Serial: {}\n".format(connection.port, connection.baud,
                                                                         connection.chain_serial)
            for board_id, board_firmware in self.gen2_addr_arr[connection.chain_serial].items():
                if board_firmware is None:
                    infos += " -> Board: 0x{:02x} Firmware: broken\n".format(board_id)
                else:
                    infos += " -> Board: 0x{:02x} Firmware: 0x{:02x}\n".format(board_id, board_firmware)

        infos += "\nIncand cards:\n"
        for incand in self.opp_incands.values():
            infos += " - Chain: {} Board: 0x{:02x} Card: {} Numbers: {}\n".format(incand.chain_serial, incand.addr,
                                                                                  incand.card_num,
                                                                                  self._get_numbers(incand.mask))

        infos += "\nInput cards:\n"
        for inputs in self.opp_inputs:
            infos += " - Chain: {} Board: 0x{:02x} Card: {} Numbers: {}\n".format(inputs.chain_serial, inputs.addr,
                                                                                  inputs.card_num,
                                                                                  self._get_numbers(inputs.mask))

        infos += "\nSolenoid cards:\n"
        for outputs in self.opp_solenoid:
            infos += " - Chain: {} Board: 0x{:02x} Card: {} Numbers: {}\n".format(outputs.chain_serial, outputs.addr,
                                                                                  outputs.card_num,
                                                                                  self._get_numbers(outputs.mask))

        infos += "\nLEDs:\n"
        for leds in self.neo_card_dict.values():
            infos += " - Chain: {} Board: 0x{:02x} Card: {}\n".format(leds.chain_serial, leds.addr, leds.card_num)

        return infos

    async def _connect_to_hardware(self):
        """Connect to each port from the config.

        This process will cause the OPPSerialCommunicator to figure out which chains they've connected to
        and to register themselves.
        """
        port_chain_serial_map = {v: k for k, v in self.config['chains'].items()}
        for port in self.config['ports']:
            # overwrite serial if defined for port
            overwrite_chain_serial = port_chain_serial_map.get(port, None)
            if overwrite_chain_serial is None and len(self.config['ports']) == 1:
                overwrite_chain_serial = port

            comm = OPPSerialCommunicator(platform=self, port=port, baud=self.config['baud'],
                                         overwrite_serial=overwrite_chain_serial)
            await comm.connect()
            self.serial_connections.add(comm)

        for chain_serial, versions in self.gen2_addr_arr.items():
            for chain_id, version in versions.items():
                if not version:
                    self.raise_config_error("Could not read version for board {}-{}.".format(chain_serial, chain_id),
                                            16)
                if self.min_version[chain_serial] != version:
                    self.raise_config_error("Version mismatch. Board {}-{} has version {:d}.{:d}.{:d}.{:d} which is not"
                                            " the minimal version "
                                            "{:d}.{:d}.{:d}.{:d}".format(chain_serial, chain_id, (version >> 24) & 0xFF,
                                                                         (version >> 16) & 0xFF, (version >> 8) & 0xFF,
                                                                         version & 0xFF,
                                                                         (self.min_version[chain_serial] >> 24) & 0xFF,
                                                                         (self.min_version[chain_serial] >> 16) & 0xFF,
                                                                         (self.min_version[chain_serial] >> 8) & 0xFF,
                                                                         self.min_version[chain_serial] & 0xFF), 1)

    def register_processor_connection(self, serial_number, communicator):
        """Register the processors to the platform.

        Args:
        ----
            serial_number: Serial number of chain.
            communicator: Instance of OPPSerialCommunicator
        """
        self.opp_connection[serial_number] = communicator

    def send_to_processor(self, chain_serial, msg):
        """Send message to processor with specific serial number.

        Args:
        ----
            chain_serial: Serial of the processor.
            msg: Message to send.
        """
        self.opp_connection[chain_serial].send(msg)

    def update_incand(self):
        """Update all the incandescents connected to OPP hardware.

        This is done once per game loop if changes have been made.

        It is currently assumed that the UART oversampling will guarantee proper
        communication with the boards.  If this does not end up being the case,
        this will be changed to update all the incandescents each loop.

        This is used for board with firmware < 2.1.0
        """
        for incand in self.opp_incands.values():
            if self.min_version[incand.chain_serial] >= 0x02010000:
                continue
            whole_msg = bytearray()
            # Check if any changes have been made
            if incand.old_state is None or (incand.old_state ^ incand.new_state) != 0:
                # Update card
                incand.old_state = incand.new_state
                msg = bytearray()
                msg.append(incand.addr)
                msg.extend(OppRs232Intf.INCAND_CMD)
                msg.extend(OppRs232Intf.INCAND_SET_ON_OFF)
                msg.append((incand.new_state >> 24) & 0xff)
                msg.append((incand.new_state >> 16) & 0xff)
                msg.append((incand.new_state >> 8) & 0xff)
                msg.append(incand.new_state & 0xff)
                msg.extend(OppRs232Intf.calc_crc8_whole_msg(msg))
                whole_msg.extend(msg)

            if whole_msg:
                # Note:  No need to send EOM at end of cmds
                send_cmd = bytes(whole_msg)

                if self.debug:
                    self.debug_log("Update incand on %s cmd:%s", incand.chain_serial,
                                   "".join(" 0x%02x" % b for b in send_cmd))
                self.send_to_processor(incand.chain_serial, send_cmd)

    @classmethod
    def get_coil_config_section(cls):
        """Return coil config section."""
        return "opp_coils"

    async def get_hw_switch_states(self):
        """Get initial hardware switch states.

        This changes switches from active low to active high
        """
        hw_states = dict()
        for opp_inp in self.opp_inputs:
            if not opp_inp.is_matrix:
                curr_bit = 1
                for index in range(0, 32):
                    if (curr_bit & opp_inp.mask) != 0:
                        if (curr_bit & opp_inp.old_state) == 0:
                            hw_states[opp_inp.chain_serial + '-' + opp_inp.card_num + '-' + str(index)] = 1
                        else:
                            hw_states[opp_inp.chain_serial + '-' + opp_inp.card_num + '-' + str(index)] = 0
                    curr_bit <<= 1
            else:
                for index in range(0, 64):
                    if ((1 << index) & opp_inp.old_state) == 0:
                        hw_states[opp_inp.chain_serial + '-' + opp_inp.card_num + '-' + str(index + 32)] = 1
                    else:
                        hw_states[opp_inp.chain_serial + '-' + opp_inp.card_num + '-' + str(index + 32)] = 0

        return hw_states

    def inv_resp(self, chain_serial, msg):
        """Parse inventory response.

        Args:
        ----
            chain_serial: Serial of the chain which received the message.
            msg: Message to parse.
        """
        self.debug_log("Received Inventory Response: %s for %s", "".join(" 0x%02x" % b for b in msg), chain_serial)

        index = 1
        self.gen2_addr_arr[chain_serial] = {}
        while msg[index] != ord(OppRs232Intf.EOM_CMD):
            if (msg[index] & ord(OppRs232Intf.CARD_ID_TYPE_MASK)) == ord(OppRs232Intf.CARD_ID_GEN2_CARD):
                self.num_gen2_brd += 1
                self.gen2_addr_arr[chain_serial][msg[index]] = None
            else:
                self.log.warning("Invalid inventory response %s for %s.", msg[index], chain_serial)
            index += 1
        self.debug_log("Found %d Gen2 OPP boards on %s.", self.num_gen2_brd, chain_serial)

    # pylint: disable-msg=too-many-statements
    @staticmethod
    def eom_resp(chain_serial, msg):
        """Process an EOM.

        Args:
        ----
            chain_serial: Serial of the chain which received the message.
            msg: Message to parse.
        """
        # An EOM command can be used to resynchronize communications if message synch is lost

    def _parse_gen2_board(self, chain_serial, msg, read_input_msg):
        has_neo = False
        has_sw_matrix = False
        has_lamp_matrix = False
        wing_index = 0
        sol_mask = 0
        inp_mask = 0
        incand_mask = 0
        while wing_index < OppRs232Intf.NUM_G2_WING_PER_BRD:
            if msg[2 + wing_index] == ord(OppRs232Intf.WING_SOL):
                sol_mask |= (0x0f << (4 * wing_index))
                inp_mask |= (0x0f << (8 * wing_index))
            elif msg[2 + wing_index] == ord(OppRs232Intf.WING_INP):
                inp_mask |= (0xff << (8 * wing_index))
            elif msg[2 + wing_index] == ord(OppRs232Intf.WING_INCAND):
                incand_mask |= (0xff << (8 * wing_index))
            elif msg[2 + wing_index] in (ord(OppRs232Intf.WING_SW_MATRIX_OUT),
                                         ord(OppRs232Intf.WING_SW_MATRIX_OUT_LOW_WING)):
                has_sw_matrix = True
            elif msg[2 + wing_index] == ord(OppRs232Intf.WING_NEO):
                has_neo = True
                inp_mask |= (0xef << (8 * wing_index))
            elif msg[2 + wing_index] == ord(OppRs232Intf.WING_HI_SIDE_INCAND):
                incand_mask |= (0xff << (8 * wing_index))
            elif msg[2 + wing_index] == ord(OppRs232Intf.WING_NEO_SOL):
                inp_mask |= (0x0e << (8 * wing_index))
                sol_mask |= (0x0f << (4 * wing_index))
                has_neo = True
            elif msg[2 + wing_index] in (ord(OppRs232Intf.WING_LAMP_MATRIX_COL_WING),
                                         ord(OppRs232Intf.WING_LAMP_MATRIX_ROW_WING)):
                has_lamp_matrix = True

            wing_index += 1
        if incand_mask != 0:
            card = OPPIncandCard(chain_serial, msg[0], incand_mask, self.machine)
            self.opp_incands["{}-{}".format(chain_serial, card.card_num)] = card
        if sol_mask != 0:
            self.opp_solenoid.append(
                OPPSolenoidCard(chain_serial, msg[0], sol_mask, self.sol_dict, self))
        if inp_mask != 0:
            # Create the input object, and add to the command to read all inputs
            self.opp_inputs.append(OPPInputCard(chain_serial, msg[0], inp_mask, self.inp_dict,
                                                self.inp_addr_dict))

            # Add command to read all inputs to read input message
            inp_msg = bytearray()
            inp_msg.append(msg[0])
            inp_msg.extend(OppRs232Intf.READ_GEN2_INP_CMD)
            inp_msg.append(0)
            inp_msg.append(0)
            inp_msg.append(0)
            inp_msg.append(0)
            inp_msg.extend(OppRs232Intf.calc_crc8_whole_msg(inp_msg))
            read_input_msg.extend(inp_msg)

        if has_sw_matrix:
            # Create the matrix object, and add to the command to read all matrix inputs
            self.opp_inputs.append(OPPMatrixCard(chain_serial, msg[0], self.inp_dict,
                                                 self.matrix_inp_addr_dict))

            # Add command to read all matrix inputs to read input message
            inp_msg = bytearray()
            inp_msg.append(msg[0])
            inp_msg.extend(OppRs232Intf.READ_MATRIX_INP)
            inp_msg.append(0)
            inp_msg.append(0)
            inp_msg.append(0)
            inp_msg.append(0)
            inp_msg.append(0)
            inp_msg.append(0)
            inp_msg.append(0)
            inp_msg.append(0)
            inp_msg.extend(OppRs232Intf.calc_crc8_whole_msg(inp_msg))
            read_input_msg.extend(inp_msg)
        if has_neo:
            card = OPPNeopixelCard(chain_serial, msg[0], self)
            self.neo_card_dict[chain_serial + '-' + card.card_num] = card
        if has_lamp_matrix:
            card = OPPModernMatrixLightsCard(chain_serial, msg[0], self)
            self.matrix_light_cards[chain_serial + '-' + card.card_num] = card

    def get_gen2_cfg_resp(self, chain_serial, msg):
        """Process cfg response.

        Args:
        ----
            chain_serial: Serial of the chain which received the message.
            msg: Message to parse.
        """
        # Multiple get gen2 cfg responses can be received at once
        self.debug_log("Received Gen2 Cfg Response:%s", "".join(" 0x%02x" % b for b in msg))
        curr_index = 0
        read_input_msg = bytearray()
        while True:
            # check that message is long enough, must include crc8
            if len(msg) < curr_index + 7:
                self.log.warning("Msg is too short: %s.", "".join(" 0x%02x" % b for b in msg))
                self.opp_connection[chain_serial].lost_synch()
                break
            # Verify the CRC8 is correct
            crc8 = OppRs232Intf.calc_crc8_part_msg(msg, curr_index, 6)
            if msg[curr_index + 6] != ord(crc8):
                self.bad_crc += 1
                self.log.warning("Msg contains bad CRC:%s.", "".join(" 0x%02x" % b for b in msg))
                break
            self._parse_gen2_board(chain_serial, msg[curr_index:curr_index + 6], read_input_msg)

            if (len(msg) > curr_index + 7) and (msg[curr_index + 7] == ord(OppRs232Intf.EOM_CMD)):
                break
            if (len(msg) > curr_index + 8) and (msg[curr_index + 8] == ord(OppRs232Intf.GET_GEN2_CFG)):
                curr_index += 7
            else:
                self.log.warning("Malformed GET_GEN2_CFG response:%s.",
                                 "".join(" 0x%02x" % b for b in msg))
                self.opp_connection[chain_serial].lost_synch()
                break

        read_input_msg.extend(OppRs232Intf.EOM_CMD)
        self.read_input_msg[chain_serial] = bytes(read_input_msg)
        self._poll_response_received[chain_serial] = asyncio.Event()
        self._poll_response_received[chain_serial].set()

    def vers_resp(self, chain_serial, msg):
        """Process version response.

        Args:
        ----
            chain_serial: Serial of the chain which received the message.
            msg: Message to parse.
        """
        # Multiple get version responses can be received at once
        self.debug_log("Received Version Response (Chain: %s): %s", chain_serial, "".join(" 0x%02x" % b for b in msg))
        curr_index = 0
        while True:
            # check that message is long enough, must include crc8
            if len(msg) < curr_index + 7:
                self.log.warning("Msg is too short (Chain: %s): %s.", chain_serial, "".join(" 0x%02x" % b for b in msg))
                self.opp_connection[chain_serial].lost_synch()
                break
            # Verify the CRC8 is correct
            crc8 = OppRs232Intf.calc_crc8_part_msg(msg, curr_index, 6)
            if msg[curr_index + 6] != ord(crc8):
                self.bad_crc += 1
                self.log.warning("Msg contains bad CRC (Chain: %s):%s.", chain_serial,
                                 "".join(" 0x%02x" % b for b in msg))
                break
            version = (msg[curr_index + 2] << 24) | \
                (msg[curr_index + 3] << 16) | \
                (msg[curr_index + 4] << 8) | \
                msg[curr_index + 5]
            self.debug_log("Firmware version of board 0x%02x (Chain: %s): %d.%d.%d.%d", msg[curr_index], chain_serial,
                           msg[curr_index + 2], msg[curr_index + 3], msg[curr_index + 4], msg[curr_index + 5])
            if msg[curr_index] not in self.gen2_addr_arr[chain_serial]:
                self.log.warning("Got firmware response for %s but not in inventory at %s", msg[curr_index],
                                 chain_serial)
            else:
                self.gen2_addr_arr[chain_serial][msg[curr_index]] = version

            if version < self.min_version[chain_serial]:
                self.min_version[chain_serial] = version
            if version == BAD_FW_VERSION:
                raise AssertionError("Original firmware sent only to Brian before adding "
                                     "real version numbers. The firmware must be updated before "
                                     "MPF will work.")
            if (len(msg) > curr_index + 7) and (msg[curr_index + 7] == ord(OppRs232Intf.EOM_CMD)):
                break
            if (len(msg) > curr_index + 8) and (msg[curr_index + 8] == ord(OppRs232Intf.GET_VERS_CMD)):
                curr_index += 7
            else:
                self.log.warning("Malformed GET_VERS_CMD response (Chain %s): %s.", chain_serial,
                                 "".join(" 0x%02x" % b for b in msg))
                self.opp_connection[chain_serial].lost_synch()
                break

    def read_gen2_inp_resp_initial(self, chain_serial, msg):
        """Read initial switch states.

        Args:
        ----
            chain_serial: Serial of the chain which received the message.
            msg: Message to parse.
        """
        # Verify the CRC8 is correct
        if len(msg) < 7:
            raise AssertionError("Received too short initial input response: " + "".join(" 0x%02x" % b for b in msg))
        crc8 = OppRs232Intf.calc_crc8_part_msg(msg, 0, 6)
        if msg[6] != ord(crc8):
            self.bad_crc += 1
            self.log.warning("Msg contains bad CRC:%s.", "".join(" 0x%02x" % b for b in msg))
        else:
            if chain_serial + '-' + str(msg[0]) not in self.inp_addr_dict:
                self.log.warning("Got input response for invalid card at initial request: %s. Msg: %s.", msg[0],
                                 "".join(" 0x%02x" % b for b in msg))
                return

            opp_inp = self.inp_addr_dict[chain_serial + '-' + str(msg[0])]
            new_state = (msg[2] << 24) | \
                (msg[3] << 16) | \
                (msg[4] << 8) | \
                msg[5]

            opp_inp.old_state = new_state

    def read_gen2_inp_resp(self, chain_serial, msg):
        """Read switch changes.

        Args:
        ----
            chain_serial: Serial of the chain which received the message.
            msg: Message to parse.
        """
        # Single read gen2 input response.  Receive function breaks them down

        # Verify the CRC8 is correct
        if len(msg) < 7:
            self.log.warning("Msg too short: %s.", "".join(" 0x%02x" % b for b in msg))
            self.opp_connection[chain_serial].lost_synch()
            return

        crc8 = OppRs232Intf.calc_crc8_part_msg(msg, 0, 6)
        if msg[6] != ord(crc8):
            self.bad_crc += 1
            self.log.warning("Msg contains bad CRC:%s.", "".join(" 0x%02x" % b for b in msg))
        else:
            if chain_serial + '-' + str(msg[0]) not in self.inp_addr_dict:
                self.log.warning("Got input response for invalid card: %s. Msg: %s.", msg[0],
                                 "".join(" 0x%02x" % b for b in msg))
                return

            opp_inp = self.inp_addr_dict[chain_serial + '-' + str(msg[0])]
            new_state = (msg[2] << 24) | \
                (msg[3] << 16) | \
                (msg[4] << 8) | \
                msg[5]

            # Update the state which holds inputs that are active
            changes = opp_inp.old_state ^ new_state
            if changes != 0:
                curr_bit = 1
                for index in range(0, 32):
                    if (curr_bit & changes) != 0:
                        if (curr_bit & new_state) == 0:
                            self.machine.switch_controller.process_switch_by_num(
                                state=1,
                                num=opp_inp.chain_serial + '-' + opp_inp.card_num + '-' + str(index),
                                platform=self)
                        else:
                            self.machine.switch_controller.process_switch_by_num(
                                state=0,
                                num=opp_inp.chain_serial + '-' + opp_inp.card_num + '-' + str(index),
                                platform=self)
                    curr_bit <<= 1
            opp_inp.old_state = new_state

        # we can continue to poll
        self._poll_response_received[chain_serial].set()

    def read_matrix_inp_resp_initial(self, chain_serial, msg):
        """Read initial matrix switch states.

        Args:
        ----
            chain_serial: Serial of the chain which received the message.
            msg: Message to parse.
        """
        # Verify the CRC8 is correct
        if len(msg) < 11:
            raise AssertionError("Received too short initial input response: " + "".join(" 0x%02x" % b for b in msg))
        crc8 = OppRs232Intf.calc_crc8_part_msg(msg, 0, 10)
        if msg[10] != ord(crc8):
            self.bad_crc += 1
            self.log.warning("Msg contains bad CRC:%s.", "".join(" 0x%02x" % b for b in msg))
        else:
            if chain_serial + '-' + str(msg[0]) not in self.matrix_inp_addr_dict:
                self.log.warning("Got input response for invalid matrix card at initial request: %s. Msg: %s.", msg[0],
                                 "".join(" 0x%02x" % b for b in msg))
                return
            opp_inp = self.matrix_inp_addr_dict[chain_serial + '-' + str(msg[0])]
            opp_inp.old_state = ((msg[2] << 56) | (msg[3] << 48) | (msg[4] << 40) | (msg[5] << 32) |
                                 (msg[6] << 24) | (msg[7] << 16) | (msg[8] << 8) | msg[9])

    # pylint: disable-msg=too-many-nested-blocks
    def read_matrix_inp_resp(self, chain_serial, msg):
        """Read matrix switch changes.

        Args:
        ----
            chain_serial: Serial of the chain which received the message.
            msg: Message to parse.
        """
        # Single read gen2 input response.  Receive function breaks them down

        # Verify the CRC8 is correct
        if len(msg) < 11:
            self.log.warning("Msg too short: %s.", "".join(" 0x%02x" % b for b in msg))
            self.opp_connection[chain_serial].lost_synch()
            return

        crc8 = OppRs232Intf.calc_crc8_part_msg(msg, 0, 10)
        if msg[10] != ord(crc8):
            self.bad_crc += 1
            self.log.warning("Msg contains bad CRC:%s.", "".join(" 0x%02x" % b for b in msg))
        else:
            if chain_serial + '-' + str(msg[0]) not in self.matrix_inp_addr_dict:
                self.log.warning("Got input response for invalid matrix card: %s. Msg: %s.", msg[0],
                                 "".join(" 0x%02x" % b for b in msg))
                return
            opp_inp = self.matrix_inp_addr_dict[chain_serial + '-' + str(msg[0])]
            new_state = ((msg[2] << 56) | (msg[3] << 48) | (msg[4] << 40) | (msg[5] << 32) |
                         (msg[6] << 24) | (msg[7] << 16) | (msg[8] << 8) | msg[9])

            changes = opp_inp.old_state ^ new_state
            if changes != 0:
                curr_bit = 1
                for index in range(32, 96):
                    if (curr_bit & changes) != 0:
                        if (curr_bit & new_state) == 0:
                            self.machine.switch_controller.process_switch_by_num(
                                state=1,
                                num=opp_inp.chain_serial + '-' + opp_inp.card_num + '-' + str(index),
                                platform=self)
                        else:
                            self.machine.switch_controller.process_switch_by_num(
                                state=0,
                                num=opp_inp.chain_serial + '-' + opp_inp.card_num + '-' + str(index),
                                platform=self)
                    curr_bit <<= 1
            opp_inp.old_state = new_state

        # we can continue to poll
        self._poll_response_received[chain_serial].set()

    def _get_dict_index(self, input_str):
        if not isinstance(input_str, str):
            self.raise_config_error("Invalid number format for OPP. Number should be card-number or chain-card-number "
                                    "(e.g. 0-1)", 2)

        try:
            chain_str, card_str, number_str = input_str.split("-")
        except ValueError:
            if len(self.serial_connections) > 1:
                self.raise_config_error("You need to specify a chain as chain-card-number in: {}".format(input_str), 17)
            else:
                chain_str = list(self.serial_connections)[0].chain_serial
            try:
                card_str, number_str = input_str.split("-")
            except ValueError:
                card_str = '0'
                number_str = input_str

        if chain_str not in self.opp_connection:
            self.raise_config_error("Chain {} does not exist. Existing chains: {}".format(
                chain_str, list(self.opp_connection.keys())), 3)

        return chain_str + "-" + card_str + "-" + number_str

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict):
        """Configure a driver.

        Args:
        ----
            config: Config dict.
            number: Number of this driver.
            platform_settings: Platform specific settings.
        """
        if not self.opp_connection:
            self.raise_config_error("A request was made to configure an OPP solenoid, "
                                    "but no OPP connection is available", 4)

        number = self._get_dict_index(number)

        if number not in self.sol_dict:
            self.raise_config_error("A request was made to configure an OPP solenoid "
                                    "with number {} which doesn't exist".format(number), 5)

        # Use new update individual solenoid command
        opp_sol = self.sol_dict[number]
        opp_sol.config = config
        opp_sol.platform_settings = platform_settings
        if self.debug:
            self.debug_log("Configure driver %s", number)
        default_pulse = PulseSettings(config.default_pulse_power, config.default_pulse_ms)
        default_hold = HoldSettings(config.default_hold_power)
        opp_sol.reconfigure_driver(default_pulse, default_hold)

        # Removing the default input is not necessary since the
        # CFG_SOL_USE_SWITCH is not being set

        return opp_sol

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict):
        """Configure a switch.

        Args:
        ----
            number: Number of this switch.
            config: Config dict.
            platform_config: Platform specific settings.
        """
        del platform_config
        del config
        # A switch is termed as an input to OPP
        if not self.opp_connection:
            self.raise_config_error("A request was made to configure an OPP switch, "
                                    "but no OPP connection is available", 6)

        number = self._get_dict_index(number)

        if number not in self.inp_dict:
            self.raise_config_error("A request was made to configure an OPP switch "
                                    "with number {} which doesn't exist".format(number), 7)

        return self.inp_dict[number]

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse number and subtype to channel."""
        if subtype in ("matrix", "incand"):
            return [
                {
                    "number": self._get_dict_index(number)
                }
            ]
        if not subtype or subtype == "led":
            full_index = self._get_dict_index(number)
            chain_serial, card, index = full_index.split('-')
            return [
                {
                    "number": "{}-{}-{}".format(chain_serial, card, int(index) * 3)
                },
                {
                    "number": "{}-{}-{}".format(chain_serial, card, int(index) * 3 + 1)
                },
                {
                    "number": "{}-{}-{}".format(chain_serial, card, int(index) * 3 + 2)
                },
            ]

        self.raise_config_error("Unknown subtype {}".format(subtype), 8)
        return []

    def configure_light(self, number, subtype, config, platform_settings):
        """Configure a led or matrix light."""
        del config
        if not self.opp_connection:
            self.raise_config_error("A request was made to configure an OPP light, "
                                    "but no OPP connection is available", 9)

        chain_serial, card, light_num = number.split('-')
        index = chain_serial + '-' + card

        if not subtype or subtype == "led":
            if index not in self.neo_card_dict:
                self.raise_config_error("A request was made to configure an OPP neopixel "
                                        "with card number {} which doesn't exist".format(card), 10)

            if not self.neo_card_dict[index].is_valid_light_number(light_num):
                self.raise_config_error("A request was made to configure an OPP neopixel "
                                        "with card number {} but number '{}' is "
                                        "invalid".format(card, light_num), 22)

            light = OPPModernLightChannel(chain_serial, int(card), int(light_num), self._light_system)
            self._light_system.mark_dirty(light)
            return light

        if subtype == "matrix" and self.min_version[chain_serial] >= 0x02010000:
            # modern matrix lights
            if index not in self.matrix_light_cards:
                self.raise_config_error("A request was made to configure an OPP matrix light "
                                        "with card number {} which doesn't exist".format(card), 18)

            if not self.matrix_light_cards[index].is_valid_light_number(light_num):
                self.raise_config_error("A request was made to configure an OPP matrix light "
                                        "with card number {} but number '{}' is "
                                        "invalid".format(card, light_num), 19)

            light = OPPModernLightChannel(chain_serial, int(card), int(light_num) + 0x2000, self._light_system)
            self._light_system.mark_dirty(light)
            return light

        if subtype in ("incand", "matrix"):
            if index not in self.opp_incands:
                self.raise_config_error("A request was made to configure an OPP incand light "
                                        "with card number {} which doesn't exist".format(card), 20)

            if not self.opp_incands[index].is_valid_light_number(light_num):
                self.raise_config_error("A request was made to configure an OPP incand light "
                                        "with card number {} but number '{}' is "
                                        "invalid".format(card, light_num), 21)
            if self.min_version[chain_serial] >= 0x02010000:
                light = self.opp_incands[index].configure_modern_fade_incand(light_num, self._light_system)
                self._light_system.mark_dirty(light)
                return light

            # legacy incands with new or old subtype
            return self.opp_incands[index].configure_software_fade_incand(light_num)

        self.raise_config_error("Unknown subtype {}".format(subtype), 12)
        return None

    async def _poll_sender(self, chain_serial):
        """Poll switches."""
        if len(self.read_input_msg[chain_serial]) <= 1:
            # there is no point in polling without switches
            return

        while True:
            # wait for previous poll response
            timeout = 1 / self.config['poll_hz'] * 25
            try:
                await asyncio.wait_for(self._poll_response_received[chain_serial].wait(), timeout)
            except asyncio.TimeoutError:
                self.log.warning("Poll took more than %sms for %s", timeout * 1000, chain_serial)
            else:
                self._poll_response_received[chain_serial].clear()
            # send poll
            self.send_to_processor(chain_serial, self.read_input_msg[chain_serial])
            await self.opp_connection[chain_serial].writer.drain()
            # the line above saturates the link and seems to overwhelm the hardware. limit it to 100Hz
            await asyncio.sleep(1 / self.config['poll_hz'])

    def _verify_coil_and_switch_fit(self, switch, coil):
        chain_serial, card, solenoid = coil.hw_driver.number.split('-')
        sw_chain_serial, sw_card, sw_num = switch.hw_switch.number.split('-')
        if self.min_version[chain_serial] >= 0x02000000:
            if chain_serial != sw_chain_serial or card != sw_card:
                self.raise_config_error('Invalid switch being configured for driver. Driver = {} '
                                        'Switch = {}. Driver and switch have to be on the same '
                                        'board.'.format(coil.hw_driver.number, switch.hw_switch.number), 13)
        else:
            matching_sw = ((int(solenoid) & 0x0c) << 1) | (int(solenoid) & 0x03)
            if chain_serial != sw_chain_serial or card != sw_card or matching_sw != int(sw_num):
                self.raise_config_error('Invalid switch being configured for driver. Driver = {} '
                                        'Switch = {}. For Firmware < 0.2.0 they have to be on the same board and '
                                        'have the same number'.format(coil.hw_driver.number, switch.hw_switch.number),
                                        14)

    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit rule on driver.

        Pulses a driver when a switch is hit. When the switch is released the pulse continues. Typically used for
        autofire coils such as pop bumpers.
        """
        self._write_hw_rule(enable_switch, coil, use_hold=False, can_cancel=False)

    def set_delayed_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings, delay_ms: int):
        """Set pulse on hit and release rule to driver.

        When a switch is hit and a certain delay passed it pulses a driver.
        When the switch is released the pulse continues.
        Typically used for kickbacks.
        """
        if delay_ms <= 0:
            raise AssertionError("set_delayed_pulse_on_hit_rule should be used with a positive delay "
                                 "not {}".format(delay_ms))
        if delay_ms > 255:
            raise AssertionError("set_delayed_pulse_on_hit_rule is limited to max 255ms "
                                 "(was {})".format(delay_ms))

        self._write_hw_rule(enable_switch, coil, use_hold=False, can_cancel=False, delay_ms=int(delay_ms))

    def set_pulse_on_hit_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and release rule to driver.

        Pulses a driver when a switch is hit. When the switch is released the pulse is canceled. Typically used on
        the main coil for dual coil flippers without eos switch.
        """
        self._write_hw_rule(enable_switch, coil, use_hold=False, can_cancel=True)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and enable and relase rule on driver.

        Pulses a driver when a switch is hit. Then enables the driver (may be with pwm). When the switch is released
        the pulse is canceled and the driver gets disabled. Typically used for single coil flippers.
        """
        self._write_hw_rule(enable_switch, coil, use_hold=True, can_cancel=True)

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                      eos_switch: SwitchSettings, coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and release and disable rule on driver.

        Pulses a driver when a switch is hit. Then enables the driver (may be with pwm). When the switch is released
        the pulse is canceled and the driver gets disabled. When the second disable_switch is hit the pulse is canceled
        and the driver gets disabled. Typically used on the main coil for dual coil flippers with eos switch.
        """
        raise AssertionError("Not implemented in OPP currently")

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver.

        Pulses a driver when a switch is hit. Then enables the driver (may be with pwm). When the switch is released
        the pulse is canceled and the driver becomes disabled. When the eos_switch is hit the pulse is canceled
        and the driver becomes enabled (likely with PWM).
        Typically used on the coil for single-wound coil flippers with eos switch.
        """
        raise AssertionError("Not implemented in OPP currently")

    # pylint: disable-msg=too-many-arguments
    def _write_hw_rule(self, switch_obj: SwitchSettings, driver_obj: DriverSettings, use_hold, can_cancel,
                       delay_ms=None):
        if switch_obj.invert:
            raise AssertionError("Cannot handle inverted switches")

        if driver_obj.hold_settings and not use_hold:
            raise AssertionError("Invalid call")

        self._verify_coil_and_switch_fit(switch_obj, driver_obj)

        self.debug_log("Setting HW Rule. Driver: %s", driver_obj.hw_driver.number)

        driver_obj.hw_driver.switches.append(switch_obj.hw_switch.number)
        driver_obj.hw_driver.set_switch_rule(driver_obj.pulse_settings, driver_obj.hold_settings, driver_obj.recycle,
                                             can_cancel, delay_ms)
        _, _, switch_num = switch_obj.hw_switch.number.split("-")
        switch_num = int(switch_num)
        self._add_switch_coil_mapping(switch_num, driver_obj.hw_driver)

    def _remove_switch_coil_mapping(self, switch_num, driver: "OPPSolenoid"):
        """Remove mapping between switch and coil."""
        if self.min_version[driver.sol_card.chain_serial] < 0x02000000:
            return

        _, _, coil_num = driver.number.split('-')
        msg = bytearray()
        msg.append(driver.sol_card.addr)
        msg.extend(OppRs232Intf.SET_SOL_INP_CMD)
        msg.append(int(switch_num))
        msg.append(int(coil_num) + ord(OppRs232Intf.CFG_SOL_INP_REMOVE))
        msg.extend(OppRs232Intf.calc_crc8_whole_msg(msg))
        msg.extend(OppRs232Intf.EOM_CMD)
        final_cmd = bytes(msg)

        if self.debug:
            self.debug_log("Unmapping input %s and coil %s on %s", switch_num, coil_num, driver.sol_card.chain_serial)
        self.send_to_processor(driver.sol_card.chain_serial, final_cmd)

    def _add_switch_coil_mapping(self, switch_num, driver: "OPPSolenoid"):
        """Add mapping between switch and coil."""
        if self.min_version[driver.sol_card.chain_serial] < 0x02000000:
            return
        _, _, coil_num = driver.number.split('-')
        msg = bytearray()
        msg.append(driver.sol_card.addr)
        msg.extend(OppRs232Intf.SET_SOL_INP_CMD)
        msg.append(int(switch_num))
        msg.append(int(coil_num))
        msg.extend(OppRs232Intf.calc_crc8_whole_msg(msg))
        msg.extend(OppRs232Intf.EOM_CMD)
        final_cmd = bytes(msg)

        if self.debug:
            self.debug_log("Mapping input %s and coil %s on %s", switch_num, coil_num, driver.sol_card.chain_serial)
        self.send_to_processor(driver.sol_card.chain_serial, final_cmd)

    def clear_hw_rule(self, switch: SwitchSettings, coil: DriverSettings):
        """Clear a hardware rule.

        This is used if you want to remove the linkage between a switch and
        some driver activity. For example, if you wanted to disable your
        flippers (so that a player pushing the flipper buttons wouldn't cause
        the flippers to flip), you'd call this method with your flipper button
        as the *sw_num*.

        """
        if switch.hw_switch.number in coil.hw_driver.switches:
            if self.debug:
                self.debug_log("Clearing HW Rule for switch: %s, coils: %s", switch.hw_switch.number,
                               coil.hw_driver.number)
            coil.hw_driver.switches.remove(switch.hw_switch.number)
            _, _, switch_num = switch.hw_switch.number.split("-")
            switch_num = int(switch_num)
            self._remove_switch_coil_mapping(switch_num, coil.hw_driver)

        # disable rule if there are no more switches
        # Technically not necessary unless the solenoid parameters are
        # changing.  MPF may not know when initial kick and hold values
        # are changed, so this might need to be called each time.
        if not coil.hw_driver.switches:
            coil.hw_driver.remove_switch_rule()



"""Contains the High Score mode code."""
import asyncio

from mpf.core.async_mode import AsyncMode
from mpf.core.player import Player


class HighScore(AsyncMode):

    """High score mode.

    Mode which runs during the game ending process to check for high scores and lets the players enter their names or
    initials.
    """

    __slots__ = ["data_manager", "high_scores", "high_score_config", "pending_award"]

    def __init__(self, *args, **kwargs):
        """Initialise high score mode."""
        self.data_manager = None
        self.high_scores = None
        self.high_score_config = None
        self.pending_award = None
        super().__init__(*args, **kwargs)

    def mode_init(self):
        """Initialise high score mode."""
        self.data_manager = self.machine.create_data_manager('high_scores')
        self.high_scores = self.data_manager.get_data()

        self.high_score_config = self.machine.config_validator.validate_config(
            config_spec='high_score',
            source=self.config.get('high_score', {}),
            section_name='high_score')

        # if data is invalid. do not use it
        if self.high_scores and not self._validate_data(self.high_scores):
            self.log.warning("High score data failed validation. Resetting to defaults.")
            self.high_scores = None

        # Load defaults if no high_scores are stored
        if not self.high_scores:
            self._load_defaults()

        self._create_machine_vars()
        self.pending_award = None

        for event in self.high_score_config['reset_high_scores_events']:
            # important: this may not be a mode handler as it should be always active
            self.machine.events.add_handler(event, self._reset)

    def _load_defaults(self):
        """Load default high scores."""
        self.high_scores = {k: [(next(iter(a.keys())), next(iter(a.values()))) for a in v] for (k, v) in
                            self.config['high_score']['defaults'].items()}

    def _reset(self, **kwargs):
        """Reset high scores."""
        del kwargs
        self._load_defaults()
        self._create_machine_vars()
        self._write_scores_to_disk()

    def _validate_data(self, data):
        try:
            for category in data:
                if category not in self.high_score_config['categories']:
                    self.log.warning("Found invalid category in high scores.")
                    return False

                for entry in data[category]:
                    if not isinstance(entry, tuple) or len(entry) != 2:
                        self.log.warning("Found invalid high score entry.")
                        return False

                    if not isinstance(entry[0], str) or not isinstance(entry[1], (int, float)):
                        self.log.warning("Found invalid data type in high score entry.")
                        return False
                if len(data[category]) != len(self.config['high_score']['defaults'][category]):
                    self.log.warning("High Score Category %s contains %s entries while defaults contain %s",
                                     category, len(data[category]),
                                     len(self.config['high_score']['defaults'][category]))
                    return False

        except TypeError:
            return False

        return True

    def _create_machine_vars(self):
        """Create all machine vars in the machine on start.

        This is used in attract mode.
        """
        for category, entries in self.high_score_config['categories'].items():
            try:
                for position, (label, (name, value)) in (
                        enumerate(zip(entries,
                                      self.high_scores[category]))):

                    self.machine.variables.set_machine_var(
                        name=category + str(position + 1) + '_label',
                        value=label)

                    '''machine_var: (high_score_category)(position)_label

                    desc: The "label" of the high score for that specific
                    score category and position. For example,
                    ``score1_label`` holds the label for the #1 position
                    of the "score" player variable (which might be "GRAND
                    CHAMPION").

                    '''

                    self.machine.variables.set_machine_var(
                        name=category + str(position + 1) + '_name',
                        value=name)

                    '''machine_var: (high_score_category)(position)_name

                    desc: Holds the player's name (or initials) for the
                    high score for that category and position.

                    '''

                    self.machine.variables.set_machine_var(
                        name=category + str(position + 1) + '_value',
                        value=value)

                    '''machine_var: (high_score_category)(position)_value

                    desc: Holds the numeric value for the high score
                    for that category and position.

                    '''

            except KeyError:
                self.high_scores[category] = list()

    # pylint: disable-msg=too-many-nested-blocks
    async def _run(self) -> None:
        """Run high score mode."""
        if not self.machine.game or not self.machine.game.player_list:
            self.warning_log("High Score started but there was no game. Will not start.")
            return

        new_high_score_list = {}

        # iterate highscore categories
        for category_name, award_names in self.high_score_config['categories'].items():
            new_list = list()

            # add the existing high scores to the list

            # make sure we have this category in the existing high scores
            if category_name in self.high_scores:
                for category_high_scores in self.high_scores[category_name]:
                    new_list.append(category_high_scores)

            # add the players scores from this game to the list
            for player in self.machine.game.player_list:
                # if the player var is 0, don't add it. This prevents
                # values of 0 being added to blank high score lists
                if player[category_name]:
                    new_list.append((player, player[category_name]))

            # sort if from highest to lowest
            new_list.sort(key=lambda x: x[1], reverse=(category_name not in self.high_score_config['reverse_sort']))

            # scan through and see if any of our players are in this list
            i = 0
            while i < len(award_names) and i < len(new_list):
                entry = new_list[i]
                if isinstance(entry[0], Player):
                    player, value = entry
                    # ask player for initials if we do not know them
                    if not player.initials:
                        try:
                            player.initials = await self._ask_player_for_initials(player, award_names[i], value)
                        except asyncio.TimeoutError:
                            del new_list[i]
                            # no entry when the player missed the timeout
                            continue
                    # add high score
                    new_list[i] = (player.initials, value)
                    # show award slide
                    await self._show_award_slide(player.initials, award_names[i], value)

                # next entry
                i += 1

            # save the new list for this category and trim it so that it's the length specified in the config
            new_high_score_list[category_name] = new_list[:len(award_names)]

        self.high_scores = new_high_score_list
        self._write_scores_to_disk()
        self._create_machine_vars()

    # pylint: disable-msg=too-many-arguments
    async def _ask_player_for_initials(self, player: Player, award_label: str, value: int) -> str:
        """Show text widget to ask player for initials."""
        self.info_log("New high score. Player: %s, award_label: %s"
                      ", Value: %s", player, award_label, value)

        self.machine.events.post('high_score_enter_initials',
                                 award=award_label,
                                 player_num=player.number,
                                 value=value)

        event_result = await asyncio.wait_for(
            self.machine.events.wait_for_event("text_input_high_score_complete"),
            timeout=self.high_score_config['enter_initials_timeout']
        )   # type: dict

        return event_result["text"] if "text" in event_result else ''

    async def _show_award_slide(self, player_name: str, award: str, value: int) -> None:
        if not self.high_score_config['award_slide_display_time']:
            return

        self.machine.events.post(
            'high_score_award_display',
            player_name=player_name,
            award=award,
            value=value)
        self.machine.events.post(
            '{}_award_display'.format(award),
            player_name=player_name,
            award=award,
            value=value)
        await asyncio.sleep(self.high_score_config['award_slide_display_time'] / 1000)

    def _write_scores_to_disk(self) -> None:
        self.data_manager.save_all(data=self.high_scores)



from mpf.tests.MpfTestCase import MpfTestCase, test_config
from unittest.mock import MagicMock


class TestBallDeviceJamSwitch(MpfTestCase):

    max_wait_ms = 200

    def get_config_file(self):
        return 'test_ball_device_jam_switch.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def _captured_from_pf(self, balls, **kwargs):
        del kwargs
        self._captured += balls

    def put_four_balls_in_trough(self):
        self._captured = 0
        self.machine.events.add_handler('balldevice_captured_from_playfield', self._captured_from_pf)
        self.machine.switch_controller.process_switch('s_trough_1', 1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_devices["trough"].balls)
        self.assertEqual(4, self._captured)
        self._captured = 0

        self.trough_coil = self.machine.coils["trough_eject"]
        self.plunger_coil = self.machine.coils["plunger_eject"]

        self.trough_coil.pulse = MagicMock()
        self.plunger_coil.pulse = MagicMock()

    @test_config("test_ball_device_jam_switch_initial.yaml")
    def test_reorder_on_startup(self):
        # test reorder on startup with a jammed trough with no ball switches active
        self.assertEqual("pulsed_2", self.machine.coils["trough_eject"].hw_driver.state)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_devices["trough"].balls)

    def test_eject_with_jam_switch(self):
        # Tests the proper operation of a trough eject with a jam switch

        self.put_four_balls_in_trough()

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # default pulse
        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(.1)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 1)
        assert not self.plunger_coil.pulse.called
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)

        # player hits the launch button
        self.machine.switch_controller.process_switch('s_launch', 1)
        self.machine.switch_controller.process_switch('s_launch', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_plunger', 0)
        self.advance_time_and_run(1)

        # ball moves from plunger lane to playfield
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.machine.switch_controller.process_switch('s_playfield', 0)
        self.advance_time_and_run(.1)

        self.plunger_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 0)
        self.assertEqual(self.machine.ball_devices["playfield"].balls, 1)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["plunger"]._state)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self._captured)

    def test_eject_no_jam_switch_activity(self):
        # Jam switch is configured, but it is not activated when the ball
        # ejects

        self.put_four_balls_in_trough()

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # default pulse
        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        # self.machine.switch_controller.process_switch('s_trough_jam', 1)
        # self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(.1)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 1)
        assert not self.plunger_coil.pulse.called
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)

        # player hits the launch button
        self.machine.switch_controller.process_switch('s_launch', 1)
        self.machine.switch_controller.process_switch('s_launch', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_plunger', 0)
        self.advance_time_and_run(1)

        # ball moves from plunger lane to playfield
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.machine.switch_controller.process_switch('s_playfield', 0)
        self.advance_time_and_run(.1)

        self.assertEqual(1, self.plunger_coil.pulse.called)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 0)
        self.assertEqual(self.machine.ball_devices["playfield"].balls, 1)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["plunger"]._state)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self._captured)

    def test_eject_with_jam_all_balls_disappear(self):
        # Ball ejects and gets stuck on jam switch. All other balls shift a
        # half position and get stuck between switches, so it's like all the
        # balls just disappeared.

        self.put_four_balls_in_trough()

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # default pulse
        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock(return_value=100)

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.advance_time_and_run(11)

        # reorder balls
        self.trough_coil.pulse.assert_called_once_with(2, max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock(return_value=100)
        self.advance_time_and_run(.5)
        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.advance_time_and_run(2)

        # soft pulse to eject only the jammed ball
        self.trough_coil.pulse.assert_called_once_with(5, max_wait_ms=self.max_wait_ms)

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.advance_time_and_run(100)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self._captured)

    def test_eject_ball_stuck_in_jam_switch(self):
        # Ball ejects, gets stuck in jam switch

        self.put_four_balls_in_trough()

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # default pulse
        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.advance_time_and_run(.1)

        # wait for timeout
        self.advance_time_and_run(10)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 4)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 0)

        # trough should retry softly
        self.trough_coil.pulse.assert_called_once_with(5, max_wait_ms=self.max_wait_ms)
        assert not self.plunger_coil.pulse.called
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(1)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.advance_time_and_run(100)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self._captured)

    def test_eject_ball_falls_back_in(self):
        # Ball ejects, ball leaves with proper timeout, but ball falls back in

        self.put_four_balls_in_trough()

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(1)

        # ball goes into plunger and comes back
        self.machine.switch_controller.process_switch('s_trough_jam', 1)

        # wait for timeout
        self.advance_time_and_run(10)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 4)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 0)
        assert not self.plunger_coil.pulse.called

        # trough should pulse softly
        self.trough_coil.pulse.assert_called_once_with(5, max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        # ball leaves and comes back again
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)

        # wait for timeout
        self.advance_time_and_run(10)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 4)
        assert not self.plunger_coil.pulse.called

        self.assertEqual(self.machine.ball_devices["plunger"].balls, 0)

        # trough should pulse softly again
        self.trough_coil.pulse.assert_called_once_with(5, max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        # ball leaves and comes back again
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)

        # wait for timeout
        self.advance_time_and_run(10)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 4)
        assert not self.plunger_coil.pulse.called

        self.assertEqual(self.machine.ball_devices["plunger"].balls, 0)

        # trough should pulse normally
        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        # ball leaves and comes back again
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)

        # wait for timeout
        self.advance_time_and_run(10)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 4)
        assert not self.plunger_coil.pulse.called

        self.assertEqual(self.machine.ball_devices["plunger"].balls, 0)

        # trough should pulse hard
        self.trough_coil.pulse.assert_called_once_with(15, max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        # ball leaves
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(1)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.advance_time_and_run(100)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self._captured)

    def test_eject_ball_stuck_with_second_ball_enter(self):
        # one ball on playfield, second ball ejects but gets stuck at jam
        # switch right as playfield ball drains

        # launch a ball into the playfield
        self.put_four_balls_in_trough()

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(.1)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 1)
        assert not self.plunger_coil.pulse.called
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)

        # player hits the launch button
        self.machine.switch_controller.process_switch('s_launch', 1)
        self.machine.switch_controller.process_switch('s_launch', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_plunger', 0)
        self.advance_time_and_run(1)

        # ball moves from plunger lane to playfield
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.machine.switch_controller.process_switch('s_playfield', 0)
        self.advance_time_and_run(.1)

        self.assertEqual(1, self.plunger_coil.pulse.called)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(1, self.machine.ball_devices["playfield"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["plunger"]._state)

        # Playfield requests a second ball
        self.machine.playfield.add_ball()
        self.advance_time_and_run(1)

        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        # ball gets stuck in jam switch while ball drains from pf
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)

        self.machine.switch_controller.process_switch('s_trough_1', 1)
        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.advance_time_and_run(.1)

        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.advance_time_and_run(10)

        # trough retries softly
        self.trough_coil.pulse.assert_called_once_with(5, max_wait_ms=self.max_wait_ms)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual(1, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.playfield.balls)
        self.assertEqual(1, self._captured)

    def test_eject_while_second_ball_enter(self):
        # one ball on playfield, second ball ejects and the other drains at the
        # same time. jam switch stays open

        # launch a ball into the playfield
        self.put_four_balls_in_trough()

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(.1)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 1)
        assert not self.plunger_coil.pulse.called
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)

        # player hits the launch button
        self.machine.switch_controller.process_switch('s_launch', 1)
        self.machine.switch_controller.process_switch('s_launch', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_plunger', 0)
        self.advance_time_and_run(1)

        # ball moves from plunger lane to playfield
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.machine.switch_controller.process_switch('s_playfield', 0)
        self.advance_time_and_run(.1)

        self.assertEqual(1, self.plunger_coil.pulse.called)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(1, self.machine.ball_devices["playfield"].balls)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["plunger"]._state)

        # Playfield requests a second ball
        self.machine.playfield.add_ball()
        self.advance_time_and_run(1)

        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        # ball gets stuck in jam switch while ball drains from pf
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)

        self.machine.switch_controller.process_switch('s_trough_1', 1)
        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.advance_time_and_run(.1)

        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(6)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 1)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual(1, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(1, self._captured)

    def test_random_jam_switch_enable(self):
        return
        # Jam switch just enables randomly

        self.put_four_balls_in_trough()

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # default pulse
        self.trough_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(.1)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 1)
        assert not self.plunger_coil.pulse.called
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)

        # player hits the launch button
        self.machine.switch_controller.process_switch('s_launch', 1)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_launch', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_plunger', 0)
        self.advance_time_and_run(1)

        # ball moves from plunger lane to playfield
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_playfield', 0)
        self.advance_time_and_run(1)

        self.plunger_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.assertEqual(3, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.ball_devices["plunger"].balls)
        self.assertEqual(1, self.machine.ball_devices["playfield"].balls)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["plunger"]._state)

        # Now the jam switch enables. This should never happen in a game, but
        # we should start to think about dealing with bad switches
        self.machine.switch_controller.process_switch('s_trough_jam', 1)

        # currently this will trigger a ball drain

        raise NotImplementedError

    def test_device_starts_with_active_jam_switch(self):
        # MPF boots with jam switch active. During eject a ball always stays
        # on the jam switch

        self.put_four_balls_in_trough()
        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.advance_time_and_run(1)

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # soft pulse
        self.trough_coil.pulse.assert_called_once_with(5, max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.advance_time_and_run(.1)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 1)
        assert not self.plunger_coil.pulse.called
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)

        # player hits the launch button
        self.machine.switch_controller.process_switch('s_launch', 1)
        self.machine.switch_controller.process_switch('s_launch', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_plunger', 0)
        self.advance_time_and_run(1)

        # ball moves from plunger lane to playfield
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.machine.switch_controller.process_switch('s_playfield', 0)
        self.advance_time_and_run(.1)

        self.plunger_coil.pulse.assert_called_once_with(max_wait_ms=self.max_wait_ms)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 0)
        self.assertEqual(self.machine.ball_devices["playfield"].balls, 1)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["plunger"]._state)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self._captured)

        # request second ball
        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # soft pulse
        self.trough_coil.pulse.assert_called_once_with(5, max_wait_ms=self.max_wait_ms)
        self.trough_coil.pulse = MagicMock()

        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.advance_time_and_run(.1)

        # ball goes missing for some time
        self.advance_time_and_run(10)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 2)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 1)
        self.assertEqual(self.machine.ball_devices["playfield"].balls, 1)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self._captured)



"""Implements the ball search procedure."""
from collections import namedtuple

from typing import List

from mpf.core.delays import DelayManager
from mpf.core.machine import MachineController
from mpf.core.mpf_controller import MpfController

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.devices.playfield import Playfield     # pylint: disable-msg=cyclic-import,unused-import

BallSearchCallback = namedtuple("BallSearchCallback", ["priority", "callback", "name", "restore_callback"])


class BallSearch(MpfController):

    """Implements Ball search for a playfield device.

    In MPF, the ball search functionality is attached to each playfield
    device, rather than being done at the global level. (In other words, each
    playfield is responsible for making sure no balls get stuck on it, and it
    leverages an instance of this BallSearch class to handle it.)
    """

    def __init__(self, machine: MachineController, playfield: "Playfield") -> None:
        """Initialize ball search."""
        self.module_name = 'BallSearch.' + playfield.name
        self.config_name = 'ball_search'

        super().__init__(machine)

        self.playfield = playfield
        """The playfield device this ball search instance is attached to."""

        self.delay = DelayManager(self.machine)

        self.started = False
        """Is the ball search process started (running) now."""
        self.enabled = False
        """Is ball search enabled."""
        self.blocked = False
        """If True, ball search will be blocked and will not start."""
        self.callbacks = []     # type: List[BallSearchCallback]

        self.iteration = False
        """Current iteration of the ball search, or ``False`` if ball search
        is not started."""
        self.iterator = False
        self.phase = False
        """Current phase of the ball search, or ``False`` if ball search is not
        started."""

        # register for events
        self.machine.events.add_handler('request_to_start_game',
                                        self.request_to_start_game)

        self.machine.events.add_handler('cancel_ball_search',
                                        self.cancel_ball_search)
        '''event: cancel_ball_search
        desc: This event will cancel all running ball searches and mark the
        balls as lost. This is only a handler so all you have to do is to post
        the event.'''

    def request_to_start_game(self, **kwargs):
        """Handle result of the *request_to_start_game* event.

        If ball search is running, this method will return *False* to prevent
        the game from starting while ball search is running.

        This method also posts the *ball_search_prevents_game_start* event
        if ball search is started.

        """
        # todo we should enable ball search if a ball is missing on game start

        del kwargs
        if self.started:

            self.machine.events.post('ball_search_prevents_game_start')
            '''event: ball_search_prevents_game_start
            desc: A game start has been requested, but the ball search process
            is running and thus the game start has been blocked. This is a
            good event to use for a slide player to inform the player that the
            machine is looking for a missing ball.'''

            return False

        return True

    def register(self, priority, callback, name, *, restore_callback=None):
        """Register a callback for sequential ball search.

        Callbacks are called by priority. Ball search only waits if the
        callback returns true.

        Args:
        ----
            priority: priority of this callback in the ball search procedure
            callback: callback to call. ball search will wait before the next
                callback, if it returns true
            name: string name which is used for debugging & the logs
            restore_callback: optional callback to restore state of the device
                after ball search ended
        """
        self.debug_log("Registering callback: {} (priority: {})".format(
            name, priority))
        self.callbacks.append(BallSearchCallback(priority, callback, name, restore_callback))
        # sort by priority
        self.callbacks = sorted(self.callbacks, key=lambda entry: entry.priority)

    def enable(self, **kwargs):
        """Enable the ball search for this playfield.

        Note that this method does *not* start the ball search process. Rather
        it just resets and starts the timeout timer, as well as resetting it
        when playfield switches are hit.

        """
        if self.blocked:
            return

        del kwargs
        if self.playfield.config['enable_ball_search'] is False or (
            not self.playfield.config['enable_ball_search'] and
                not self.machine.config['mpf']['default_ball_search']):
            return

        if not self.callbacks:
            raise AssertionError("No callbacks registered")

        self.debug_log("Enabling Ball Search")

        self.enabled = True

        self.reset_timer()

    def disable(self, **kwargs):
        """Disable ball search.

        This method will also stop the ball search if it is running.
        """
        del kwargs
        self.stop()

        self.debug_log("Disabling Ball Search")
        self.enabled = False
        self.delay.remove('start')

    def block(self, **kwargs):
        """Block ball search for this playfield.

        Blocking will disable ball search if it's enabled or running, and will
        prevent ball search from enabling if it's disabled until
        ``ball_search_unblock()`` is called.
        """
        del kwargs
        self.debug_log("Blocking ball search")
        self.disable()
        self.blocked = True

    def unblock(self, **kwargs):
        """Unblock ball search for this playfield.

        This will check to see if there are balls on the playfield, and if so,
        enable ball search.
        """
        del kwargs
        self.debug_log("Unblocking ball search")
        self.blocked = False

        if self.playfield.balls:
            self.enable()

    def reset_timer(self):
        """Reset the timeout timer which starts ball search.

        This method will also cancel an actively running (started) ball search.

        This is called by the playfield anytime a playfield switch is hit.

        """
        if self.started:
            self.stop()

        if self.enabled:
            self.debug_log("Resetting ball search timer")
            self.delay.reset(name='start', callback=self.start,
                             ms=self.playfield.config['ball_search_timeout'])

    def start(self):
        """Start ball search the ball search process."""
        if not self.enabled or self.started or not self.callbacks:
            return
        self.started = True
        self.iteration = 1
        self.phase = 1
        self.iterator = iter(self.callbacks)
        self.info_log("Starting ball search")
        self.machine.events.post('ball_search_started')
        '''event: ball_search_started

        desc: The ball search process has been begun.
        '''

        self.machine.events.post('ball_search_phase_1', iteration=1)
        # see description below

        self._run()

    def stop(self):
        """Stop an actively running ball search."""
        if not self.started:
            return

        self.info_log("Stopping ball search")

        self.started = False
        self.delay.remove('run')

        # restore all devices
        for callback in self.callbacks:
            if callback.restore_callback:
                callback.restore_callback()

        self.machine.events.post('ball_search_stopped')
        '''event: ball_search_stopped

        desc: The ball search process has been disabled. This event is posted
            any time ball search stops, regardless of whether it found a ball
            or gave up. (If the ball search failed to find the ball, it will
            also post the *ball_search_failed* event.)
        '''

    def _run(self):
        # Runs one iteration of the ball search.
        # Will schedule itself for the next run.

        # check if we should skip this phase
        if not self.playfield.config['ball_search_phase_{}_searches'.format(self.phase)]:
            self.phase += 1
            if self.phase > 3:
                # give up
                self.give_up()
            else:
                # go to the next phase
                self._run()
            return

        timeout = self.playfield.config['ball_search_interval']

        # iterate until we are done with all callbacks
        while True:
            try:
                element = next(self.iterator)
            except StopIteration:
                self.iteration += 1
                self.machine.events.post('ball_search_phase_{}'.format(self.phase),
                                         iteration=self.iteration)
                '''event: ball_search_phase_(num)

                desc: The ball search phase (num) has started.
                args:
                    iteration: Current iteration of phase (num)
                '''
                # give up at some point
                if self.iteration > self.playfield.config[
                        'ball_search_phase_{}_searches'.format(self.phase)]:
                    self.phase += 1
                    self.iteration = 1
                    if self.phase > 3:
                        self.give_up()
                        return

                self.iterator = iter(self.callbacks)
                element = next(self.iterator)
                timeout = self.playfield.config[
                    'ball_search_wait_after_iteration']

            # if a callback returns True we wait for the next one
            self.debug_log("Ball search: {} (phase: {}  iteration: {})".format(
                element.name, self.phase, self.iteration))
            if element.callback(self.phase, self.iteration):
                self.delay.add(name='run', callback=self._run, ms=timeout)
                return

    def cancel_ball_search(self, **kwargs):
        """Cancel the current ball search and mark the ball as missing."""
        del kwargs
        if self.started:
            self.give_up()

    def give_up(self):
        """Give up the ball search.

        This method is called when the ball search process Did not find the
        missing ball. It executes the failed action which depending on the specification of *ball_search_failed_action*,
        either adds a replacement ball, ends the game, or ends the current ball.
        """
        self.info_log("Ball Search failed to find ball. Giving up!")
        self.disable()
        self.machine.events.post('ball_search_failed')
        '''event: ball_search_failed

        desc: The ball search process has failed to locate a missing or stuck
            ball and has given up. This event will be posted immediately after
            the *ball_search_stopped* event.
        '''

        lost_balls = self.playfield.balls
        self.machine.ball_controller.num_balls_known -= lost_balls
        self.playfield.balls = 0
        self.playfield.available_balls = 0

        self._compensate_lost_balls(lost_balls)

    def _compensate_lost_balls(self, lost_balls):
        if not self.machine.game:
            return

        if self.playfield.config['ball_search_failed_action'] == "new_ball":
            if self.machine.ball_controller.num_balls_known > 0:
                # we have at least one ball remaining
                self.info_log("Adding %s replacement ball", lost_balls)
                for dummy_iterator in range(lost_balls):
                    self.playfield.add_ball()
            else:
                self.info_log("No more balls left. Ending game!")
                self.machine.game.end_game()

        elif self.playfield.config['ball_search_failed_action'] == "end_game":
            if self.machine.game:
                self.info_log("Ending the game")
                self.machine.game.end_game()
            else:
                self.warning_log("There is no game. Doing nothing!")

        elif self.playfield.config['ball_search_failed_action'] == "end_ball":
            self.info_log("Ending current ball")
            self.machine.game.end_ball()

        else:
            raise AssertionError("Unknown action " + self.playfield.config[
                'ball_search_failed_action'])



import asyncio

from mpf.platforms.pololu import pololu_tic
from mpf.tests.MpfTestCase import MpfTestCase
import ruamel.yaml


class TestPololuTic(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/pololu_tic/'

    def get_platform(self):
        return False

    def _ticcmd_async(self, *args):
        if args == ("-s", "--full"):
            return ruamel.yaml.dump({
                'Errors currently stopping the motor': [],
                'Current position': self._position
            })
        elif args == ('--reset-command-timeout',):
            return ""

        if args not in self.expected_commands:
            raise AssertionError("Unexpected command {}. Expected: {}".format(args, self.expected_commands))

        return_value = self.expected_commands[args]
        del self.expected_commands[args]
        return return_value

    def setUp(self):
        self._position = 0
        self.expected_commands = {
            ('--energize',): "",
            ('--exit-safe-start',): "",
            ('--current', '192'): "",
            ('--max-decel', '40000'): "",
            ('--max-accel', '40000'): "",
            ('--starting-speed', '0'): "",
            ('--max-speed', '2000000'): "",
            ('--step-mode', '1'): "",
            ('--velocity', '2000000'): ""
        }
        def _create_fake_thread(self_inner):
            self_inner.stop_future = asyncio.Future(loop=self.loop)
            self_inner.loop = self.loop

        def _stop_thread(self_inner):
            pass

        pololu_tic.PololuTiccmdWrapper._start_thread = _create_fake_thread
        pololu_tic.PololuTiccmdWrapper._stop_thread = _stop_thread
        pololu_tic.PololuTiccmdWrapper._run_subprocess_ticcmd = self._ticcmd_async

        super().setUp()
        self.assertFalse(self.expected_commands)

    def test_tic(self):
        """Test Tic stepper."""
        stepper = self.machine.steppers["stepper1"]

        # stepper arrives at home
        self.expected_commands = {
            ('--halt-and-hold',): "",
            ('--halt-and-set-position', '0'): ""
        }
        self.hit_switch_and_run("s_home", 1)
        self.assertFalse(self.expected_commands)

        # move stepper
        self.expected_commands = {
            ('--position', '20'): ""
        }
        self.post_event("test_01")
        self.advance_time_and_run(.1)
        self.assertFalse(self.expected_commands)
        self.assertEqual(20, stepper._target_position)
        self.assertEqual(0, stepper._current_position)

        # stepper arrives at position
        self._position = 20
        self.advance_time_and_run(.1)
        self.assertEqual(20, stepper._target_position)
        self.assertEqual(20, stepper._current_position)

        # move stepper
        self.expected_commands = {
            ('--position', '50'): ""
        }
        self.post_event("test_10")
        self.advance_time_and_run(.1)
        self.assertFalse(self.expected_commands)
        self.assertEqual(50, stepper._target_position)
        self.assertEqual(20, stepper._current_position)

        # stepper arrives at position
        self._position = 50
        self.advance_time_and_run(.1)
        self.assertEqual(50, stepper._target_position)
        self.assertEqual(50, stepper._current_position)

        # prevent crash during shutdown
        self.expected_commands = {
            ('--halt-and-hold',): ""
        }
